Package: eisaR
Title: Exon-Intron Split Analysis (EISA) in R
Version: 1.25.0
Authors@R: c(person("Michael", "Stadler", email = "michael.stadler@fmi.ch", role = c("aut", "cre")),
             person("Dimos", "Gaidatzis", email = "dimosthenis.gaidatzis@fmi.ch", role = "aut"),
             person("Lukas", "Burger", email = "lukas.burger@fmi.ch", role = "aut"),
             person("Charlotte", "Soneson", email = "charlotte.soneson@fmi.ch", role = "aut"))
Description: Exon-intron split analysis (EISA) uses ordinary RNA-seq data to measure
    changes in mature RNA and pre-mRNA reads across different experimental conditions
    to quantify transcriptional and post-transcriptional regulation of gene expression.
    For details see Gaidatzis et al., Nat Biotechnol 2015. doi: 10.1038/nbt.3269.
    eisaR implements the major steps of EISA in R. 
Depends: R (>= 4.1)
License: GPL-3
biocViews: Transcription, GeneExpression, GeneRegulation,
        FunctionalGenomics, Transcriptomics, Regression, RNASeq
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.3.1
Suggests: knitr, rmarkdown, testthat, BiocStyle, QuasR, Rbowtie,
        Rhisat2, Biostrings, BSgenome, BSgenome.Hsapiens.UCSC.hg38,
        ensembldb, AnnotationDbi, GenomicFeatures, txdbmaker,
        rtracklayer, withr
VignetteBuilder: knitr
Imports: graphics, stats, GenomicRanges, S4Vectors, IRanges, limma,
        edgeR (>= 4.0), methods, SummarizedExperiment, BiocGenerics,
        utils
URL: https://github.com/fmicompbio/eisaR
BugReports: https://github.com/fmicompbio/eisaR/issues
git_url: https://git.bioconductor.org/packages/eisaR
git_branch: devel
git_last_commit: 939131c
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 22:29:33 UTC; biocbuild
Author: Michael Stadler [aut, cre],
  Dimos Gaidatzis [aut],
  Lukas Burger [aut],
  Charlotte Soneson [aut]
Maintainer: Michael Stadler <michael.stadler@fmi.ch>

# Generated by roxygen2: do not edit by hand

export(exportToGtf)
export(getFeatureRanges)
export(getRegionsFromTxDb)
export(getTx2Gene)
export(plotEISA)
export(runEISA)
import(GenomicRanges)
import(edgeR)
importFrom(BiocGenerics,end)
importFrom(BiocGenerics,relist)
importFrom(BiocGenerics,setdiff)
importFrom(BiocGenerics,sort)
importFrom(BiocGenerics,start)
importFrom(BiocGenerics,subset)
importFrom(BiocGenerics,unlist)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,reduce)
importFrom(IRanges,IRanges)
importFrom(S4Vectors,elementNROWS)
importFrom(S4Vectors,mcols)
importFrom(S4Vectors,metadata)
importFrom(S4Vectors,runLength)
importFrom(S4Vectors,split)
importFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(SummarizedExperiment,assay)
importFrom(SummarizedExperiment,assayNames)
importFrom(graphics,legend)
importFrom(graphics,plot)
importFrom(graphics,points)
importFrom(limma,nonEstimable)
importFrom(methods,is)
importFrom(stats,model.matrix)
importFrom(utils,write.table)

<img src="inst/www/eisaR.png" align="right" alt="" width="120" />

# `eisaR`: Exon-Intron Split Analysis (EISA) in R

<br>

## Overview

Exon-intron split analysis (`EISA`) uses ordinary RNA-seq data to measure
changes in mature RNA and pre-mRNA reads across different experimental
conditions to quantify transcriptional and post-transcriptional regulation
of gene expression.

For details see Gaidatzis et al., Nat Biotechnol 2015. doi: 10.1038/nbt.3269.  
`eisaR` implements the major steps of EISA in R. 
In addition, it contains functionality for extracting spliced and unspliced 
transcript sequences, as well as intron sequences (with similar options as 
the [BUSpaRse](https://github.com/BUStools/BUSpaRse)) package), from an 
annotated genome. These sequences can be indexed and used, e.g., for 
quantification in preparation for RNA velocity estimation.

Developed by:

- [Michael Stadler](https://github.com/mbstadler)
- Dimos Gaidatzis
- [Lukas Burger](https://github.com/LukasBurger)
- [Charlotte Soneson](https://github.com/csoneson)

Also a big "thank you" for contributions to:

- [Federico Marini](https://github.com/federicomarini)

## Installation

To install `eisaR` from Bioconductor, you will need at least R 4.0 and Bioconductor 3.11, which is available on April 28, 2020. Then use the following within R (see also [eisaR download page](https://bioconductor.org/packages/eisaR/)):
```
# BiocManager is needed to install Bioconductor packages
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# Install eisaR
BiocManager::install("eisaR")
```

## Functionality

All you need is RNA-seq data from at least two conditions (e.g. wildtype and
mutant). The `eisaR` package contains convenience functions to facilitate the
steps in an exon-intron split analysis, which consists of:  

1. preparing the annotation (exonic and gene body coordinate ranges)  
2. quantifying RNA-seq alignments in exons and introns  
3. calculating and comparing exonic and intronic changes across conditions  
4. visualizing the results  

For the steps 1. and 2. above, this `eisaR` vignette makes use of Bioconductor
annotation and the [QuasR](https://bioconductor.org/packages/QuasR/) package.
It is also possible to obtain count tables for exons and introns using some
other pipeline or approach, and directly start with step 3.


## Reference
`EISA` has been described in:  

"Analysis of intronic and exonic reads in RNA-seq data characterizes
transcriptional and post-transcriptional regulation."  
Gaidatzis D., Burger L., Florescu M. and Stadler, M.B.  
*Nat Biotechnol.* **2015**; 33(7):722-9.
[PubMed: 26098447](https://www.ncbi.nlm.nih.gov/pubmed/26098447), [doi: 10.1038/nbt.3269](https://doi.org/10.1038/nbt.3269)

The functionality for extracting spliced and unspliced transcript sequences,
as well as intron sequences, for use in RNA velocity analysis, has been described in:

"Preprocessing choices affect RNA velocity results for droplet scRNA-seq data."
Soneson C., Srivastava A., Patro R. and Stadler, M.B.
*PLoS Comput Biol* **2021**; 17(1):e1008585.
[PubMed: 33428615](https://pubmed.ncbi.nlm.nih.gov/33428615/), [doi: 10.1371/journal.pcbi.1008585](https://doi.org/10.1371/journal.pcbi.1008585)


## Software status

| Platforms        |  OS              | R CMD check      | Coverage         | 
|:----------------:|:----------------:|:----------------:|:----------------:|
| GitHub Actions | Linux/Windows/macOS | [![R build status](https://github.com/fmicompbio/eisaR/workflows/R-CMD-check/badge.svg)](https://github.com/fmicompbio/eisaR/actions) | [![Codecov.io coverage status](https://codecov.io/github/fmicompbio/eisaR/coverage.svg?branch=master)](https://codecov.io/github/fmicompbio/eisaR) |
| Bioc ([_devel_](http://bioconductor.org/packages/devel/bioc/html/eisaR.html)) | Multiple | [![Bioconductor-devel Build Status](http://bioconductor.org/shields/build/devel/bioc/eisaR.svg)](http://bioconductor.org/checkResults/devel/bioc-LATEST/eisaR) | [![Codecov.io coverage status](https://codecov.io/github/fmicompbio/eisaR/coverage.svg?branch=master)](https://codecov.io/github/fmicompbio/eisaR) |
| Bioc ([_release_](http://bioconductor.org/packages/release/bioc/html/eisaR.html)) | Multiple | [![Bioconductor-release Build Status](http://bioconductor.org/shields/build/release/bioc/eisaR.svg)](http://bioconductor.org/checkResults/release/bioc-LATEST/eisaR) | [![Codecov.io coverage status](https://codecov.io/github/fmicompbio/eisaR/coverage.svg?branch=master)](https://codecov.io/github/fmicompbio/eisaR) |

#' @title Run Exon-Intron Split Analysis.
#'
#' @description Starting from count tables with exonic and intronic counts
#'   for two conditions, perform all the steps in EISA (normalize, identify
#'   quantifyable genes, calculate contrasts and their significance).
#'
#' @author Michael Stadler
#'
#' @param cntEx Gene by sample \code{matrix} with exonic counts, OR a
#'   \code{SummarizedExperiment} with two assays named \code{exon} and
#'   \code{intron}, containing exonic and intronic counts, respectively. If
#'   \code{cntEx} is a \code{SummarizedExperiment}, \code{cntIn} will be
#'   disregarded.
#' @param cntIn Gene by sample \code{matrix} with intronic counts. Must have the
#'   same structure as \code{cntEx} (same number and order of rows and columns)
#'   if \code{cntEx} is a matrix. Will be disregarded if \code{cntEx} is a
#'   \code{SummarizedExperiment}.
#' @param cond \code{numeric}, \code{character} or \code{factor} with two levels
#'   that groups the samples (columns of \code{cntEx} and \code{cntIn}) into two
#'   conditions. The contrast will be defined as secondLevel - firstLevel.
#' @param method One of \code{NULL} (the default) or \code{"Gaidatzis2015"}. If
#'   \code{"Gaidatzis2015"}, gene filtering, statistical analysis and
#'   calculation of contrasts is performed as described in Gaidatzis et al.
#'   2015, and the statistical analysis is based on \code{\link[edgeR]{glmFit}}
#'   and \code{\link[edgeR:glmFit]{glmLRT}}. This is done by setting the arguments
#'   \code{modelSamples}, \code{geneSelection}, \code{effects}, \code{pscnt},
#'   \code{statFramework}, \code{sizeFactor}, \code{recalcNormFactAfterFilt}
#'   and \code{recalcLibSizeAfterFilt}
#'   to appropriate values (see details), overriding the defaults or any value
#'   passed to these arguments. If \code{NULL}, the default values of the arguments
#'   will be used instead (recommended).
#' @param modelSamples Whether to include a sample identifier in the design matrix
#'   of the statistical model. If \code{TRUE}, potential sample effects
#'   that affect both exonic and intronic counts of that sample will be taken
#'   into account, which could result in higher sensitivity (default: \code{TRUE}).
#' @param geneSelection Controls how to select quantifyable genes. One of the
#'   following:\describe{
#'       \item{\code{"filterByExpr"}: }{(default) First, counts are normalized using
#'       \code{\link[edgeR]{calcNormFactors}}, treating intronic and exonic counts
#'       as individual samples. Then, \code{\link[edgeR]{filterByExpr}} is used
#'       with default parameters to select quantifyable genes.}
#'       \item{\code{"none"}: }{This will use all the genes provided in the count
#'       tables, assuming that an appropriate selection of quantifyable genes has
#'       already been done.}
#'       \item{\code{"Gaidatzis2015"}: }{First, intronic and exonic counts are
#'       linearly scaled to the mean library size (estimated as the sum of all
#'       intronic or exonic counts, respectively). Then, quantifyable genes are
#'       selected as the genes with counts \code{x} that fulfill
#'       \code{log2(x + 8) > 5} in both exons and introns.}
#'   }
#' @param statFramework Selects the framework within \code{edgeR} that is used
#'   for the statistical analysis. One of:\describe{
#'       \item{\code{"QLF"}: }{(default) Quasi-likelihood F-test using
#'       \code{\link[edgeR:glmQLFTest]{glmQLFit}} and \code{\link[edgeR]{glmQLFTest}}. This
#'       framework is highly recommended as it gives stricter error rate control
#'       by accounting for the uncertainty in dispersion estimation.}
#'       \item{\code{"LRT"}: }{Likelihood ratio test using \code{\link[edgeR]{glmFit}}
#'       and \code{\link[edgeR:glmFit]{glmLRT}}}.
#'   }
#' @param legacyQLF Whether to use the 'legacy' version of
#'   \code{\link[edgeR:glmQLFTest]{glmQLFit}}. See \code{\link[edgeR:glmQLFTest]{glmQLFit}}
#'   for more details. If \code{FALSE}, the new method introduced in
#'   \code{edgeR} 4.0.0 is used.
#' @param effects How the effects (contrasts or log2 fold-changes) are calculated.
#'   One of:\describe{
#'       \item{\code{"predFC"}: }{(default) Fold-changes are calculated using
#'       the fitted model with \code{\link[edgeR]{predFC}} with
#'       \code{prior.count = pscnt}. Please note that if a sample factor is
#'       included in the model (\code{modelSamples=TRUE}), effects cannot be
#'       obtained from that model. In that case, effects are obtained from a
#'       simpler model without sample effects.}
#'       \item{\code{"Gaidatzis2015"}: }{Fold-changes are calculated using the
#'       formula \code{log2((x + pscnt)/(y + pscnt))}. If \code{pscnt} is not
#'       set to 8, \code{runEISA} will warn that this deviates from the method
#'       used in Gaidatzis et al., 2015.}
#'   }
#' @param pscnt \code{numeric(1)} with pseudocount to add to read counts
#'   (default: 2). For \code{method = "Gaidatzis2015"}, it is set to 8.
#'   It is added to scaled read counts used in \code{geneSelection = "Gaidatzis2015"}
#'   and \code{effects = "Gaidatzis2015"}, or else used in \code{cpm(..., prior.count = pscnt)}
#'   and \code{predFC(..., prior.count = pscnt)}.
#' @param sizeFactor How the size factors are calculated in the analysis.
#'   If 'exon' (default), the exon-derived size factors are used also for the
#'   columns corresponding to intronic counts. If 'intron', the intron-derived
#'   size factors are used also for the columns corresponding to exonic
#'   counts. If 'individual', column-wise size factors are calculated.
#' @param recalcNormFactAfterFilt Logical, indicating whether normalization
#'   factors should be recalculated after filtering out lowly
#'   expressed genes.
#' @param recalcLibSizeAfterFilt Logical, indicating whether library
#'   sizes should be recalculated after filtering out lowly
#'   expressed genes.
#' @param ... additional arguments passed to the \code{\link[edgeR]{DGEList}}
#'   constructor, such as \code{lib.size} or \code{genes}.
#'
#' @details Setting \code{method = "Gaidatzis2015"} has precedence over other
#'   argument values and corresponds to setting:
#'   \code{modelSamples = FALSE, geneSelection = "Gaidatzis2015",
#'   statFramework = "LRT", effects = "Gaidatzis2015", pscnt = 8,
#'   sizeFactor = "individual", recalcNormFactAfterFilt = TRUE,
#'   recalcLibSizeAfterFilt = FALSE}.
#'
#' @return a \code{list} with elements \describe{
#'   \item{fracIn}{fraction intronic counts in each sample}
#'   \item{contrastName}{contrast name}
#'   \item{contrasts}{contrast matrix for quantifyable genes, with average log2
#'     fold-changes in exons (\code{Dex}), in introns (\code{Din}), and average
#'     difference between log2 fold-changes in exons and introns (\code{Dex.Din})}
#'   \item{DGEList}{\code{\link[edgeR]{DGEList}} object used in model fitting}
#'   \item{tab.ExIn}{statisical results for differential changes between exonic
#'   and intronic contrast, an indication for post-transcriptional regulation.}
#'   \item{contr.ExIn}{contrast vector used for testing the difference between
#'   exonic and intronic contrast (results in \code{tab.ExIn})}
#'   \item{designMatrix}{design matrix used for testing the difference between
#'   exonic and intronic contrast (results in \code{tab.ExIn})}
#'   \item{params}{a \code{list} with parameter values used to run EISA}
#' }
#'
#' @references Analysis of intronic and exonic reads in RNA-seq data characterizes
#'   transcriptional and post-transcriptional regulation.
#'   Dimos Gaidatzis, Lukas Burger, Maria Florescu and Michael B. Stadler
#'   Nature Biotechnology, 2015 Jul;33(7):722-9. doi: 10.1038/nbt.3269.
#'
#' @seealso \code{\link[edgeR]{DGEList}} for \code{DGEList} object construction,
#'   \code{\link[edgeR]{calcNormFactors}} for normalization,
#'   \code{\link[edgeR]{filterByExpr}} for gene selection,
#'   \code{\link[edgeR]{glmFit}} and \code{\link[edgeR:glmQLFTest]{glmQLFit}} for statistical
#'   analysis.
#'
#' @examples
#' cntEx <- readRDS(system.file("extdata", "Fig3abc_GSE33252_rawcounts_exonic.rds",
#'                              package = "eisaR"))[,-1]
#' cntIn <- readRDS(system.file("extdata", "Fig3abc_GSE33252_rawcounts_intronic.rds",
#'                              package = "eisaR"))[,-1]
#' cond <- factor(c("ES","ES","TN","TN"))
#' res <- runEISA(cntEx, cntIn, cond)
#' plotEISA(res)
#'
#' @import edgeR
#' @importFrom limma nonEstimable
#' @importFrom stats model.matrix
#' @importFrom methods is
#' @importFrom SummarizedExperiment assay assayNames SummarizedExperiment
#'
#' @export
runEISA <- function(cntEx, cntIn, cond, method = NULL,
                    modelSamples = TRUE,
                    geneSelection = c("filterByExpr", "none", "Gaidatzis2015"),
                    statFramework = c("QLF", "LRT"),
                    legacyQLF = FALSE,
                    effects = c("predFC", "Gaidatzis2015"),
                    pscnt = 2L,
                    sizeFactor = c("exon", "intron", "individual"),
                    recalcNormFactAfterFilt = TRUE,
                    recalcLibSizeAfterFilt = FALSE, ...) {
    # check arguments
    # ... count matrices
    if (is(cntEx, "SummarizedExperiment")) {
        if (all(c("exon", "intron") %in% SummarizedExperiment::assayNames(cntEx))) {
            cntIn <- SummarizedExperiment::assay(cntEx, "intron")
            cntEx <- SummarizedExperiment::assay(cntEx, "exon")
        } else if (all(c("spliced", "unspliced") %in% SummarizedExperiment::assayNames(cntEx))) {
            cntIn <- SummarizedExperiment::assay(cntEx, "unspliced")
            cntEx <- SummarizedExperiment::assay(cntEx, "spliced")
        } else {
            stop("'cntEx' needs to have assayNames 'intron'/'exon' or 'unspliced'/'spliced'.")
        }
    }
    if (is.data.frame(cntEx))
        cntEx <- as.matrix(cntEx)
    if (is.data.frame(cntIn))
        cntIn <- as.matrix(cntIn)
    # ... consistency between cntEx and cntIn
    stopifnot(exprs = {
        is.matrix(cntEx)
        is.matrix(cntIn)
        identical(dim(cntEx), dim(cntIn))
    })
    nsmpls <- ncol(cntEx)
    if (is.null(rownames(cntEx)))
        rownames(cntEx) <- as.character(seq.int(nrow(cntEx)))
    if (is.null(colnames(cntEx)))
        colnames(cntEx) <- as.character(seq.int(nsmpls))
    if (is.null(rownames(cntIn)))
        rownames(cntIn) <- as.character(seq.int(nrow(cntIn)))
    if (is.null(colnames(cntIn)))
        colnames(cntIn) <- as.character(seq.int(ncol(cntIn)))
    stopifnot(identical(dimnames(cntEx), dimnames(cntIn)))
    # ... conditions
    if (is.numeric(cond) || is.character(cond))
        cond <- factor(cond, levels = unique(cond))
    # ... valid arguments
    geneSelection <- match.arg(geneSelection)
    statFramework <- match.arg(statFramework)
    effects <- match.arg(effects)
    sizeFactor <- match.arg(sizeFactor)
    stopifnot(exprs = {
        # cond
        is.factor(cond)
        nlevels(cond) == 2L
        length(cond) == nsmpls
        # method
        is.null(method) || identical(method, "Gaidatzis2015")
        # modelSamples
        is.logical(modelSamples)
        length(modelSamples) == 1L
        # pscnt
        is.numeric(pscnt)
        length(pscnt) == 1L
    })

    # override arguments for Gaidatzis2015
    if (!is.null(method) && method == "Gaidatzis2015") {
        message("setting parameters according to Gaidatzis et al., 2015")
        modelSamples <- FALSE
        geneSelection <- "Gaidatzis2015"
        statFramework <- "LRT"
        effects <- "Gaidatzis2015"
        pscnt <- 8L
        sizeFactor <- "individual"
        recalcNormFactAfterFilt <- TRUE
        recalcLibSizeAfterFilt <- FALSE
    }

    # fraction intronic
    fracIn <- colSums(cntIn) / (colSums(cntEx) + colSums(cntIn))

    # create DGEList
    # first nsmpls columns = exons, last nsmpls columns = introns
    cnt <- data.frame(Ex = cntEx, In = cntIn)
    y <- edgeR::DGEList(counts = cnt, ...)

    # calculate normalization factors and library sizes based on exons only, or
    # based on the individual columns
    y$samples$norm.factors.exons <- rep(edgeR::calcNormFactors(cntEx), 2L)
    y$samples$norm.factors.introns <- rep(edgeR::calcNormFactors(cntIn), 2L)
    y$samples$norm.factors.individual <- edgeR::calcNormFactors(y$counts)
    y$samples$lib.size.exons <- rep(colSums(cntEx), 2L)
    y$samples$lib.size.introns <- rep(colSums(cntIn), 2L)
    y$samples$lib.size.individual <- colSums(y$counts)
    # y <- edgeR::calcNormFactors(y)

    # create design matrix
    cond2 <- rep(cond, 2L)
    region <- factor(rep(c("ex", "in"), each = nsmpls),
                     levels = c("in", "ex"))
    smpl <- factor(rep(sprintf("s%03d", seq.int(nsmpls)), 2L))
    if (modelSamples) {
        dsgn <- model.matrix(~ smpl)
        c1Ex <- cond2 == levels(cond2)[1L] & region == "ex"
        c2Ex <- cond2 == levels(cond2)[2L] & region == "ex"
        dsgn <- cbind(dsgn, c1Ex, c2Ex)
        # dsgn <- model.matrix(~ smpl + region * cond2)
        # # need to remove a coefficient to make the design full rank
        # toRemove <- limma::nonEstimable(dsgn)
        # dsgn <- dsgn[, -match(toRemove, colnames(dsgn))]
    } else {
        dsgn <- model.matrix(~ region * cond2)
    }
    rownames(dsgn) <- colnames(cnt)

    # identify quantifyable genes and calculate log-expression values
    # here, each column is normalized with its own size factors
    y$samples$norm.factors <- y$samples$norm.factors.individual
    y$samples$lib.size <- y$samples$lib.size.individual
    if (geneSelection == "none") {
        message("skip filtering for quantifyable genes")
        logcpmEx <- edgeR::cpm(y[, seq.int(nsmpls)], log = TRUE,
                               prior.count = pscnt)
        logcpmIn <- edgeR::cpm(y[, nsmpls + seq.int(nsmpls)], log = TRUE,
                               prior.count = pscnt)
    } else {
        message("filtering quantifyable genes...", appendLF = FALSE)

        if (geneSelection == "filterByExpr") {
            quantGenes <- rownames(cntEx)[
                edgeR::filterByExpr(y[, seq.int(nsmpls)],
                                    design = dsgn[seq.int(nsmpls), ]) &
                    edgeR::filterByExpr(y[, nsmpls + seq.int(nsmpls)],
                                        design = dsgn[nsmpls + seq.int(nsmpls), ])
            ]
            logcpmEx <- edgeR::cpm(y[, seq.int(nsmpls)], log = TRUE,
                                   prior.count = pscnt)
            logcpmIn <- edgeR::cpm(y[, nsmpls + seq.int(nsmpls)], log = TRUE,
                                   prior.count = pscnt)

        } else if (geneSelection == "Gaidatzis2015") {
            # scale counts to the mean library size separately for exons and introns
            normcntEx <- t(t(cntEx) / colSums(cntEx) * mean(colSums(cntEx)))
            normcntIn <- t(t(cntIn) / colSums(cntIn) * mean(colSums(cntIn)))

            # log transform (add pseudocount)
            if (pscnt != 8L) {
                warning("Using a 'pscnt' different from 8 deviates ",
                        "from geneSelection='Gaidatzis2015'")
            }
            logcpmEx <- log2(normcntEx + pscnt)
            logcpmIn <- log2(normcntIn + pscnt)

            # Identify quantifyable genes
            quantGenes <- rownames(cntEx)[rowMeans(logcpmEx) > 5.0 &
                                              rowMeans(logcpmIn) > 5.0]

        }
        message("keeping ", length(quantGenes), " from ", nrow(y), " (",
                round(length(quantGenes) * 100.0 / nrow(y), 1L), "%)")
        y <- y[quantGenes, ]

        # Recalculate normalization factors after filtering out lowly
        # expressed genes
        if (recalcNormFactAfterFilt) {
            y$samples$norm.factors.exons <- rep(
                edgeR::calcNormFactors(y$counts[, seq.int(nsmpls)]), 2L
            )
            y$samples$norm.factors.introns <- rep(
                edgeR::calcNormFactors(y$counts[, nsmpls + seq.int(nsmpls)]), 2L
            )
            y$samples$norm.factors.individual <- edgeR::calcNormFactors(y$counts)
        }
        if (recalcLibSizeAfterFilt) {
            y$samples$lib.size.exons <- rep(
                colSums(y$counts[, seq.int(nsmpls)]), 2L
            )
            y$samples$lib.size.introns <- rep(
                colSums(y$counts[, nsmpls + seq.int(nsmpls)]), 2L
            )
            y$samples$lib.size.individual <- colSums(y$counts)
        }
        # y <- edgeR::calcNormFactors(y)
        logcpmEx <- logcpmEx[quantGenes, ]
        logcpmIn <- logcpmIn[quantGenes, ]
    }

    # statistical analysis
    if (sizeFactor == "exon") {
        y$samples$lib.size <- y$samples$lib.size.exons
        y$samples$norm.factors <- y$samples$norm.factors.exons
    } else if (sizeFactor == "intron") {
        y$samples$lib.size <- y$samples$lib.size.introns
        y$samples$norm.factors <- y$samples$norm.factors.introns
    } else {
        y$samples$lib.size <- y$samples$lib.size.individual
        y$samples$norm.factors <- y$samples$norm.factors.individual
    }
    if (any(table(cond) < 2L)) {
        warning("Need at least two replicates per condition to perform ",
                "statistical analysis. 'ExIn' result will be empty.")
        ttExIn <- list(table = data.frame())
        contr <- NULL
    } else {
        message("fitting statistical model...", appendLF = FALSE)
        y <- edgeR::estimateDisp(y, dsgn)
        if (modelSamples) {
            contr <- (colnames(dsgn) == "c2Ex") -
                (colnames(dsgn) == "c1Ex")
        } else {
            contr <- as.numeric(colnames(dsgn) == colnames(dsgn)[ncol(dsgn)])
        }
        if (statFramework == "QLF") {
            fit <- edgeR::glmQLFit(y, dsgn, legacy = legacyQLF)
            tstExIn <- edgeR::glmQLFTest(fit, contrast = contr)
        } else if (statFramework == "LRT") {
            fit <- edgeR::glmFit(y, dsgn)
            tstExIn <- edgeR::glmLRT(fit, contrast = contr)
        }
        ttExIn <- edgeR::topTags(tstExIn, n = nrow(y), sort.by = "none")
        message("done")
    }

    # calculate log-fold changes
    message("calculating log-fold changes...", appendLF = FALSE)
    contrastName <- paste(levels(cond)[2L], "-", levels(cond)[1L])
    if (effects == "predFC") {
        if (is.null(y$common.dispersion))
            stop("effects='predFC' requires a fitted model - rerun with effects='Gaidatzis2015'")
        lfc <- edgeR::predFC(y, dsgn, prior.count = pscnt)
        if (modelSamples) {
            rownames(lfc) <- rownames(y)
            dIn <- rowMeans(lfc[, colnames(lfc) %in% paste0("smpl", smpl[cond2 == levels(cond2)[2L]]), drop = FALSE]) -
                rowMeans(lfc[, colnames(lfc) %in% paste0("smpl", smpl[cond2 == levels(cond2)[1L]]), drop = FALSE])
            dExMinusDIn <- lfc[, "c2Ex"] - lfc[, "c1Ex"]
            dEx <- dIn + dExMinusDIn
            # lfc2 <- lfc
            # message("fitting model without sample factor...", appendLF = FALSE)
            # dsgn2 <- model.matrix(~ region * cond2)
            # y2 <- edgeR::estimateDisp(y, dsgn2)
            # if (statFramework == "QLF") {
            #     fit2 <- edgeR::glmQLFit(y2, dsgn2, legacy = legacyQLF)
            # } else if (statFramework == "LRT") {
            #     fit2 <- edgeR::glmFit(y2, dsgn2)
            # }
            # lfc2 <- edgeR::predFC(y2, dsgn2, prior.count = pscnt)
        } else {
            rownames(lfc) <- rownames(y)
            dEx <- rowSums(lfc[, c(3L, 4L)])
            dIn <- lfc[, 3L]
            dExMinusDIn <- lfc[, ncol(lfc)]
            # lfc2 <- lfc
        }
        # rownames(lfc) <- rownames(lfc2) <- rownames(y)
        # dEx <- rowSums(lfc2[, c(3, 4)])
        # dIn <- lfc2[, 3]
        # dExMinusDIn <- lfc[, ncol(lfc)]
        # remark: for modelSamples=TRUE, should the interaction effect be estimated...
        #         - from the simpler model (as the condition effects, for consistency among effects)
        #         - from the full model (for consistency with the interaction FDR and topTags table)
        #           -> current implementation
    } else if (effects == "Gaidatzis2015") {
        i1 <- which(cond == levels(cond)[1L])
        i2 <- which(cond == levels(cond)[2L])
        dEx <- rowMeans(logcpmEx[, i2, drop = FALSE]) -
            rowMeans(logcpmEx[, i1, drop = FALSE])
        dIn <- rowMeans(logcpmIn[, i2, drop = FALSE]) -
            rowMeans(logcpmIn[, i1, drop = FALSE])
        dExMinusDIn <- dEx - dIn
    }
    message("done")

    ## return results
    return(list(fracIn = fracIn,
                contrastName = contrastName,
                contrasts = cbind(Dex = dEx, Din = dIn, Dex.Din = dExMinusDIn),
                DGEList = y,
                tab.ExIn = ttExIn$table,
                contr.ExIn = contr,
                designMatrix = dsgn,
                params = list(method = method, modelSamples = modelSamples,
                              geneSelection = geneSelection,
                              statFramework = statFramework,
                              effects = effects, pscnt = pscnt,
                              sizeFactor = sizeFactor)))
}

test_that("plotEISA() runs", {
    cntEx <- readRDS(system.file("extdata",
                                 "Fig3abc_GSE33252_rawcounts_exonic.rds",
                                 package = "eisaR"))[, -1L]
    cntIn <- readRDS(system.file("extdata",
                                 "Fig3abc_GSE33252_rawcounts_intronic.rds",
                                 package = "eisaR"))[, -1L]
    cond <- factor(c("ES", "ES", "TN", "TN"))
    res1 <- runEISA(cntEx, cntIn, cond, method = "Gaidatzis2015")
    expect_warning(res2 <- runEISA(cntEx[, c(1L, 3L)], cntIn[, c(1L, 3L)],
                                   cond[c(1L, 3L)], method = "Gaidatzis2015"))

    tf <- tempfile(fileext = ".pdf")
    pdf(file = tf)

    expect_null(plotEISA(res1, col = "purple"))
    expect_null(plotEISA(res1, contrast = "none"))
    expect_error(plotEISA(res2))

    dev.off()
    unlink(tf)
})

context("runEISA runs")
test_that("runEISA() runs", {
    cntEx <- readRDS(system.file("extdata", "Fig3abc_GSE33252_rawcounts_exonic.rds",
                                 package = "eisaR"))[, -1L]
    cntIn <- readRDS(system.file("extdata", "Fig3abc_GSE33252_rawcounts_intronic.rds",
                                 package = "eisaR"))[, -1L]
    cond <- factor(c("ES", "ES", "TN", "TN"))
    cntSE1 <- SummarizedExperiment::SummarizedExperiment(
        assays = list(exon = cntEx, intron = cntIn)
    )
    cntSE2 <- SummarizedExperiment::SummarizedExperiment(
        assays = list(spliced = cntEx, unspliced = cntIn)
    )

    # arguments
    expect_error(runEISA(data.frame("a")))
    expect_error(runEISA(rbind(1L:2L, 3L:4L), "b"))
    expect_error(runEISA(rbind(1L:2L, 3L:4L), rbind(1L:2L, 3L:4L), c("a", "b", "c")))
    expect_warning(runEISA(rbind(1L:2L, 3L:4L), rbind(1L:2L, 3L:4L), c("a", "b"),
                           effects = "Gaidatzis2015"))
    expect_warning(runEISA(cntEx, cntIn, cond, geneSelection = "Gaidatzis2015", pscnt = 4L))
    expect_error(runEISA(SummarizedExperiment(assays = list(exon = cntEx))))

    # expected results
    res0 <- runEISA(cntEx[seq.int(1000L), ], cntIn[seq.int(1000L), ], cond,
                    modelSamples = TRUE, geneSelection = "none", statFramework = "LRT")
    res0se <- runEISA(cntEx = cntSE2[seq.int(1000L), ], cntIn = NULL, cond,
                      modelSamples = TRUE, geneSelection = "none", legacyQLF = TRUE)
    res1 <- runEISA(cntEx, cntIn, cond, method = "Gaidatzis2015")
    res1se <- runEISA(cntEx = cntSE1, cntIn = NULL, cond, method = "Gaidatzis2015")
    res2 <- runEISA(cntEx, cntIn, cond, method = NULL, modelSamples = FALSE,
                    sizeFactor = "individual", legacyQLF = TRUE)
    res3 <- runEISA(cntEx, cntIn, cond, recalcLibSizeAfterFilt = TRUE,
                    sizeFactor = "intron", legacyQLF = TRUE)
    expect_is(res0, "list")
    expect_is(res0se, "list")
    expect_identical(res0$contrasts, res0se$contrasts)
    expect_is(res1, "list")
    expect_is(res1se, "list")
    expect_identical(res1, res1se)
    expect_is(res2, "list")
    expect_is(res3, "list")
    expect_length(res1, 8L)
    expect_true(all(rownames(res1$DGEList) %in% rownames(cntEx)))
    ids <- intersect(rownames(res1$DGEList), rownames(res2$DGEList))
    expect_gt(cor(res1$contrasts[ids, "Dex"], res2$contrasts[ids, "Dex"]), 0.99)
    expect_gt(cor(res1$contrasts[ids, "Din"], res2$contrasts[ids, "Din"]), 0.99)
    expect_gt(cor(res1$contrasts[ids, "Dex.Din"], res2$contrasts[ids, "Dex.Din"]), 0.97)
    expect_gt(cor(res2$contrasts[ids, "Dex"], res3$contrasts[ids, "Dex"]), 0.99)

    # one replicate per condition
    expect_warning(res1 <- runEISA(cntEx[, c(1L, 3L)], cntIn[, c(1L, 3L)],
                                   cond[c(1L, 3L)], method = "Gaidatzis2015"))
    expect_is(res1, "list")
    expect_length(res1, 8L)
    expect_named(res1, c("fracIn", "contrastName", "contrasts", "DGEList",
                         "tab.ExIn", "contr.ExIn", "designMatrix", "params"))
    expect_is(res1$tab.ExIn, "data.frame")
    expect_identical(nrow(res1$tab.ExIn), 0L)
    expect_error(plotEISA(res1))
    expect_error(suppressWarnings(runEISA(cntEx[, c(1L, 3L)], cntIn[, c(1L, 3L)],
                                          cond[c(1L, 3L)], method = NULL,
                                          legacyQLF = TRUE)))
})

context("runEISA gives expected results")
test_that("runEISA() gives expected results", {
    # construct artificial example
    fracIntron <- 0.2
    ngenes <- 3000L
    nsig <- 80L
    nsig2 <- 20L
    set.seed(5L)

    # ... 2 conditions (ES, TN),
    #     2 replicates (exon counts, constant library size, just sampling noise)
    cond <- factor(c("ES", "ES", "TN", "TN"))
    tmp <- readRDS(system.file("extdata", "Fig3abc_GSE33252_rawcounts_exonic.rds",
                               package = "eisaR"))[, -1L]
    cntEx <- cbind(rmultinom(2L, round(sum(tmp) / 4.0), rowMeans(tmp[, 1L:2L])),
                   rmultinom(2L, round(sum(tmp) / 4.0), rowMeans(tmp[, 3L:4L])))
    colnames(cntEx) <- colnames(tmp)

    # ... select ngenes expressed genes
    selgenes <- sample(x = which(rowMeans(log2(cntEx + 1L)) > 5.0),
                       size = ngenes, replace = FALSE)
    cntEx <- cntEx[selgenes, ]

    # ... intron counts as sub-sampled exon counts with fracIntron total counts
    nEx <- colSums(cntEx)
    cntIn <- cbind(rmultinom(2L, round(sum(nEx[1L:2L]) / 2.0 * fracIntron),
                             rowMeans(cntEx[, 1L:2L])),
                   rmultinom(2L, round(sum(nEx[3L:4L]) / 2.0 * fracIntron),
                             rowMeans(cntEx[, 3L:4L])))
    dimnames(cntIn) <- dimnames(cntEx)

    # ... nsig genes with significant a interaction
    #     - multiplicative effect, in order to keep the n_exon/N_total
    #       and n_intron/N_total ratios the same
    #     - half applied to ES and NP, each (in order to get both directions)
    selsig <- rownames(cntEx)[sample(x = nrow(cntEx), size = nsig, replace = FALSE)]
    lfcsig <- sample(x = 2L:5L, size = nsig, replace = TRUE)
    names(lfcsig) <- selsig
    i <- seq(1L, round(nsig / 2.0))
    cntEx[selsig[i],  1L:2L] <- round(cntEx[selsig[i],  1L:2L] * 2.0^lfcsig[i])
    cntEx[selsig[-i], 3L:4L] <- round(cntEx[selsig[-i], 3L:4L] * 2.0^lfcsig[-i])

    # ... nsig2 genes with a sample specific effect
    #     - re-adjust counts of one replicate only, both for exons and introns
    #       (keeps the effect constant, but gives a sample-specific offset)
    selsig2 <- sample(x = selsig, size = nsig2, replace = FALSE)
    cntEx[selsig2, 4L] <- round(cntEx[selsig2, 4L] / 2.0^(lfcsig[match(selsig2, selsig)] / 2.0))
    cntIn[selsig2, 4L] <- round(cntIn[selsig2, 4L] / 2.0^(lfcsig[match(selsig2, selsig)] / 2.0))

    # run EISA (use gene filtering that is independent of model)
    res1 <- runEISA(cntEx, cntIn, cond, geneSelection = "Gaidatzis2015",
                    pscnt = 8L, sizeFactor = "individual", modelSamples = FALSE,
                    legacyQLF = TRUE)
    res2 <- runEISA(cntEx, cntIn, cond, geneSelection = "Gaidatzis2015",
                    pscnt = 8L, sizeFactor = "individual", modelSamples = TRUE,
                    legacyQLF = TRUE)

    # account for filtered genes
    ngenes <- nrow(res1$tab.ExIn)
    selsig <- intersect(selsig, rownames(res1$tab.ExIn))
    nsig <- length(selsig)
    selsig2 <- intersect(selsig2, selsig)
    nsig2 <- length(selsig2)
    lfcsig <- lfcsig[selsig]

    # look at results
    # plot(res1$tab.ExIn[, "FDR"], res2$tab.ExIn[, "FDR"], log = "xy",
    #      col = (rownames(res1$tab.ExIn) %in% selsig) + (rownames(res1$tab.ExIn) %in% selsig2) + 1)
    # abline(a=0, b=1)
    # plot(res1$contrasts[selsig, "Dex.Din"], res2$contrasts[selsig, "Dex.Din"],
    #      col = (selsig %in% selsig2) + 1)
    # abline(a=0, b=1)
    # plotEISA(res1)
    # plotEISA(res2)
    t1 <- table(sample = rownames(res1$tab.ExIn) %in% selsig2,
                true.sig = rownames(res1$tab.ExIn) %in% selsig,
                test.sig = res1$tab.ExIn$FDR < 0.01) # close to 1/nsig -> one false positive
    t2 <- table(sample = rownames(res2$tab.ExIn) %in% selsig2,
                true.sig = rownames(res2$tab.ExIn) %in% selsig,
                test.sig = res2$tab.ExIn$FDR < 0.01) # close to 1/nsig -> one false positive

    # no false positives or false negatives (except for selsig2 genes in t1)
    # these are now missed after updated df calculation in limma
    # (https://code.bioconductor.org/browse/limma/commit/e8ff47777e7c25d6dcb9582d9fcb8f9c4fed9da1)
    nmissed1 <- t1["FALSE", "TRUE", "FALSE"]
    # expect to miss some genes with sample-specific effects
    nmissed2 <- t1["TRUE", "TRUE", "FALSE"]
    expect_identical(as.vector(t1),
                     c(ngenes - nsig, 0L, nmissed1, nmissed2, 0L, 0L,
                       nsig - nsig2 - nmissed1, nsig2 - nmissed2))
    expect_identical(as.vector(t2),
                     c(ngenes - nsig, 0L, 0L, 0L, 0L, 0L, nsig - nsig2, nsig2))
})

#' @title Get exonic/gene body regions from a transcript database.
#'
#' @description From a transcript database package (\code{\link[GenomicFeatures:TxDb-class]{TxDb}}),
#'   extract exonic and gene body ranges for use with EISA. These regions can
#'   be used to quantify RNA-seq alignments in exons and gene bodies, respectively.
#'   Intronic counts can then be obtained from the difference between gene bodies
#'   and exonic region counts.
#'
#' @author Michael Stadler
#'
#' @param txdb a \code{TxDb} or an \code{EnsDb} object with the transcript annotations.
#' @param exonExt \code{numeric} (default = 10L). Exonic ranges will be extended
#'   on either side by this many nucleotides, in order to avoid "bleed-over" of
#'   exonic alignments into adjacent intronic regions.
#' @param strandedData \code{logical(1)}. If \code{TRUE}, the RNA-seq data is
#'   assumed to be strand-specific, and therefore only overlapping genes that
#'   are on the same strand will be filtered out. If \code{FALSE}, also genes
#'   overlapping on opposite strands will be filtered out.
#'
#' @details The exonic regions are generated as follows:
#'   \enumerate{
#'     \item extract exons by gene from the \code{txdb}
#'     \item extend each exon by \code{exonExt}
#'     \item combine overlapping exons within each gene
#'     \item create gene body ranges from the most extreme exonic coordinates
#'     \item filter out genes that have only a single exon (no intron), have exons
#'           on more than a single chromosome or on both strands, or that
#'           overlap other genes
#'     }
#' @return a \code{list} with elements "exons" and "genebodies", containing
#'   named \code{GenomicRanges} objects with ranges for exons and gene bodies,
#'   respectively.
#'
#' @seealso \code{\link[GenomicFeatures:TxDb-class]{TxDb}} for details on
#'   \code{TxDb} objects and the \code{txdbmaker} package for how to create them,
#'   e.g. from \code{.gtf} files.
#'
#' @examples
#' if (requireNamespace("AnnotationDbi", quietly = TRUE)) {
#'     txdb <- AnnotationDbi::loadDb(system.file("extdata", "hg19sub.sqlite", package = "eisaR"))
#'     regL <- getRegionsFromTxDb(txdb)
#'     lengths(regL)
#' }
#'
#' @import GenomicRanges
#' @importFrom S4Vectors elementNROWS runLength
#' @importFrom IRanges IRanges
#'
#' @export
getRegionsFromTxDb <- function(txdb, exonExt = 10L, strandedData = TRUE) {
    # check arguments
    stopifnot(exprs = {
        inherits(txdb, "TxDb") || inherits(txdb, "EnsDb")
        is.numeric(exonExt)
        length(exonExt) == 1L
        is.logical(strandedData)
        length(strandedData) == 1L
    })
    if (!requireNamespace("GenomicFeatures", quietly = TRUE)) {
        stop("getRegionsFromTxDb() requires installing the Bioconductor package 'GenomicFeatures'",
             " using BiocManager::install(\"GenomicFeatures\")")
    }

    # get exons and extend by exonExt on both sides
    message("extracting exon coordinates")
    exL <- GenomicFeatures::exonsBy(txdb, by = "gene")
    suppressWarnings(exL <- exL + exonExt) # may contain out of chromosome regions
    # remark: should GenomicRanges::trim now, but there is not trim,GRangesList method
    #         will suppressWarnings below until I can use trim,GRanges
    exL <- GenomicRanges::reduce(exL) # fuse

    # identify genes with single exon
    nEx <- S4Vectors::elementNROWS(exL)
    severalExons <- nEx > 1L

    # identify genes with exons on multiple chromosomes
    suppressWarnings(nChr <- S4Vectors::elementNROWS(
        S4Vectors::runLength(GenomicRanges::seqnames(exL))
    ))
    singleChr <- nChr == 1L

    ## identify genes with exons on multiple strands
    suppressWarnings(nStr <- S4Vectors::elementNROWS(
        S4Vectors::runLength(GenomicRanges::strand(exL))
    ))
    singleStrand <- nStr == 1L

    # get gene body
    ex <- suppressWarnings(GenomicRanges::trim(unlist(exL)))
    suppressWarnings(gbody <- unlist(range(exL))[unique(names(ex))])

    # trim out-of chromosome regions
    gbody <- GenomicRanges::trim(gbody)

    ## identify overlapping genes
    nov <- GenomicRanges::countOverlaps(gbody, gbody, ignore.strand = !strandedData)
    noOverlaps <- nov == 1L # only self-overlaps

    ## filter regions
    sel <- severalExons & singleChr & singleStrand & noOverlaps
    selex <- names(ex) %in% names(gbody[sel])
    message("total number of genes/exons: ", length(sel), "/", length(ex))
    message("removing overlapping/single-exon/ambiguous genes (", length(sel) - sum(sel), ")")
    message("creating filtered regions for ", sum(sel), " genes (", round(mean(sel) * 100.0, 1L),
            "%) with ", sum(selex), " exons (", round(mean(selex) * 100.0, 1L), "%)")
    gbody <- gbody[sel]
    ex <- ex[selex]

    ## return results
    return(list(exons = ex, genebodies = gbody))
}

test_that("getRegionsFromTxDb() runs", {
    requireNamespace("txdbmaker")
    requireNamespace("GenomicFeatures")
    txdb <- AnnotationDbi::loadDb(system.file("extdata", "hg19sub.sqlite", package = "eisaR"))
    regL <- getRegionsFromTxDb(txdb)

    # arguments
    expect_error(getRegionsFromTxDb(seq.int(10L)))
    expect_error(getRegionsFromTxDb(txdb, exonExt = "a"))
    expect_error(getRegionsFromTxDb(txdb, strandedData = "a"))

    # expected results
    expect_type(regL, "list")
    expect_length(regL, 2L)
    expect_equivalent(lengths(regL), c(20L, 4L))
    expect_true(all(countOverlaps(regL$exons) == 1L))

    # same for EnsDb object
    # ... load GRanges with annotation (Y)
    load(system.file("YGRanges.RData", package = "ensembldb"))
    # ... create EnsDb
    suppressWarnings(eDB <- ensembldb::ensDbFromGRanges(Y, path = tempdir(), version = 75L,
                                                        organism = "Homo_sapiens"))
    edb <- ensembldb::EnsDb(eDB)
    # ... create corresponding TxDb
    suppressWarnings(txdb <- txdbmaker::makeTxDbFromGRanges(Y))
    # ... compare annotation
    expect_length(GenomicFeatures::genes(edb),
                  length(GenomicFeatures::genes(txdb)))
    expect_length(GenomicFeatures::exons(edb),
                  length(GenomicFeatures::exons(txdb)))
    # ... compare results
    regL1 <- getRegionsFromTxDb(edb)
    regL2 <- getRegionsFromTxDb(txdb)
    expect_identical(regL1, regL2)

    # clean up
    AnnotationDbi::dbFileDisconnect(AnnotationDbi::dbconn(txdb))
})

test_that("getRegionsFromTbx() fails when required packages are missing", {
    # these tests assume that all non-base packages are installed in
    # .Library.site and will fail if this is not the case (e.g. on BioC builders)
    skip_on_bioc()
    # also skip on r-universe (https://docs.r-universe.dev/publish/troubleshoot-build.html#how-to-know-whether-tests-are-run-on-r-universe)
    skip_if_not(identical(Sys.getenv("MY_UNIVERSE"), ""))

    # set new lib paths
    # ... unload ns
    unloadNamespace("ensembldb")
    unloadNamespace("txdbmaker")
    unloadNamespace("GenomicFeatures")
    withr::with_temp_libpaths({
        dummy <- list()
        class(dummy) <- "TxDb"
        expect_error(getRegionsFromTxDb(dummy))
    }, action = "replace")
    # ... reload ns
    requireNamespace("GenomicFeatures")
    requireNamespace("txdbmaker")
    requireNamespace("ensembldb")
})

#' Generate a GRangesList object with feature ranges
#'
#' Generate a GRangesList object with genomic ranges for (any combination of)
#' spliced transcripts, unspliced transcripts and introns.
#'
#' @param gtf Path to gtf file.
#' @param featureType Character vector indicating the type(s) of features to
#'   extract, any subset of \code{c("spliced", "intron", "unspliced")}.
#' @param intronType Character vector indicating how to define the introns
#'   (only used if "intron" is part of \code{featureType}). Has to be either
#'   "separate" (introns are defined for each transcript separately) or
#'   "collapse" (transcripts of the same gene are first collapsed before
#'   introns are defined as any non-exonic part of the gene locus).
#' @param flankLength Integer scalar indicating the length of the flanking
#'   sequence added to each side of each extracted intron (only used if
#'   "intron" is included in \code{featureType}).
#' @param joinOverlappingIntrons Logical scalar indicating whether two introns
#'   that overlap (after adding the flanking sequences) should be joined into
#'   one feature.
#' @param collapseIntronsByGene Logical scalar indicating whether to collapse
#'   overlapping intron ranges by gene after extraction.
#' @param keepIntronInFeature Logical scalar indicating whether introns
#'   (after adding the flank length) should be restricted to stay within the
#'   boundaries of the feature they were generated from. For backwards
#'   compatibility, the default is \code{FALSE}.
#' @param verbose Logical scalar, whether to print out progress messages.
#'
#' @author Charlotte Soneson
#'
#' @export
#'
#' @return Returns a \code{GRangesList} object where each element represents
#'   one extracted feature. The metadata of this object contains two
#'   \code{data.frame}s mapping corresponding identifiers between the
#'   different feature types, as well as a list of all features for each type.
#'
#' @importFrom GenomicRanges GRangesList reduce
#' @importFrom BiocGenerics unlist relist setdiff start end
#' @importFrom S4Vectors mcols metadata
#'
#' @examples
#'   ## Get feature ranges
#'   grl <- getFeatureRanges(
#'     gtf = system.file("extdata/small_example.gtf", package = "eisaR"),
#'     featureType = c("spliced", "intron"),
#'     intronType = "separate",
#'     flankLength = 5L,
#'     joinOverlappingIntrons = FALSE,
#'     collapseIntronsByGene = FALSE,
#'     verbose = TRUE
#'   )
#'
#'   ## GRangesList
#'   grl
#'
#'   ## Corresponding transcript/gene IDs
#'   S4Vectors::metadata(grl)$corrtx
#'   S4Vectors::metadata(grl)$corrgene
#'
#'   ## List of features of different types
#'   S4Vectors::metadata(grl)$featurelist
#'
#'   ## Get feature sequences
#'   if (requireNamespace("BSgenome", quietly = TRUE) &&
#'       requireNamespace("GenomicFeatures", quietly = TRUE)) {
#'     library(BSgenome)
#'     genome <- Biostrings::readDNAStringSet(
#'       system.file("extdata/small_example_genome.fa", package = "eisaR"))
#'     seqs <- GenomicFeatures::extractTranscriptSeqs(x = genome,
#'                                                    transcripts = grl)
#'     seqs
#'   }
#'
getFeatureRanges <- function(gtf,
                             featureType = c("spliced", "intron"),
                             intronType = "separate",
                             flankLength = 90L,
                             joinOverlappingIntrons = FALSE,
                             collapseIntronsByGene = FALSE,
                             keepIntronInFeature = FALSE,
                             verbose = TRUE) {
    ## --------------------------------------------------------------------- ##
    ## Pre-flight checks
    ## --------------------------------------------------------------------- ##
    for (pkg in c("GenomicFeatures", "AnnotationDbi", "txdbmaker")) {
        if (!requireNamespace(pkg, quietly = TRUE)) {
            stop("getFeatureRanges() requires installing the Bioconductor package '",
                 pkg, "' using BiocManager::install(\"", pkg, "\")")
        }
    }
    if (length(gtf) != 1L || !is.character(gtf) || !file.exists(gtf)) {
        stop("'gtf' must be a character scalar providing ",
             "the path to an existing file.")
    }
    if (!all(is.character(featureType)) ||
        !all(featureType %in% c("spliced", "unspliced", "intron"))) {
        stop("'featureType' must be a subset of c('spliced',",
             "'unspliced', 'intron')")
    }
    if ("intron" %in% featureType) {
        if (length(intronType) != 1L || !is.character(intronType) ||
            !(intronType %in% c("separate", "collapse"))) {
            stop("'intronType' must be either 'separate' or 'collapse'")
        }
        if (!is.numeric(flankLength) || length(flankLength) != 1L ||
            flankLength < 0L) {
            stop("'flankLength' must be a numeric non-negative scalar")
        }
        flankLength <- as.integer(flankLength)
        if (!is.logical(joinOverlappingIntrons) ||
            length(joinOverlappingIntrons) != 1L) {
            stop("'joinOverlappingIntrons' must be a logical scalar")
        }
        if (!is.logical(collapseIntronsByGene) ||
            length(collapseIntronsByGene) != 1L) {
            stop("'collapseIntronsByGene' must be a logical scalar")
        }
        if (!is.logical(keepIntronInFeature) ||
            length(keepIntronInFeature) != 1L) {
            stop("'keepIntronInFeature' must be a logical scalar")
        }
    }
    if (!is.logical(verbose) || length(verbose) != 1L) {
        stop("'verbose' must be a logical scalar")
    }

    ## --------------------------------------------------------------------- ##
    ## Define suffixes for unspliced transcripts and introns
    ## --------------------------------------------------------------------- ##
    suffixes <- c(intron = "-I", unspliced = "-U")

    ## --------------------------------------------------------------------- ##
    ## Construct TxDb from gtf file
    ## --------------------------------------------------------------------- ##
    # suppress the warning:
    #    "genome version information is not available for this TxDb object"
    suppressWarnings(
        txdb <- txdbmaker::makeTxDbFromGFF(gtf, format = "gtf", chrominfo = NULL)
    )

    ## Initialize GRangesList
    grlfull <- GenomicRanges::GRangesList()
    featurelist <- list()

    ## --------------------------------------------------------------------- ##
    ## Group exons by transcript/gene
    ## --------------------------------------------------------------------- ##
    ## Group exons by transcript. When using exonsBy with by = "tx",
    ## the returned exons are ordered by ascending rank for each transcript,
    ## that is, by their position in the transcript.
    ebt <- GenomicFeatures::exonsBy(txdb, by = "tx", use.names = TRUE)
    t2g <- AnnotationDbi::select(txdb, keys = names(ebt),
                                 keytype = "TXNAME", columns = "GENEID")
    e2 <- BiocGenerics::unlist(ebt)
    e2$transcript_id <- names(e2)
    e2$gene_id <- t2g$GENEID[match(e2$transcript_id, t2g$TXNAME)]
    e2$exon_id <- e2$exon_name
    e2$exon_name <- NULL
    e2$type <- "exon"
    names(e2) <- NULL
    mcols(e2) <- S4Vectors::mcols(e2)[, c("exon_id", "exon_rank",
                                          "transcript_id", "gene_id", "type")]
    ebt <- BiocGenerics::relist(e2, ebt)

    ## Group exons by gene
    ebg <- GenomicFeatures::exonsBy(txdb, by = "gene")

    corrtx <- data.frame(spliced = unique(names(ebt)),
                         stringsAsFactors = FALSE)
    corrtx$unspliced <- paste0(corrtx$spliced, suffixes["unspliced"])
    corrgene <- data.frame(spliced = unique(unlist(ebt)$gene_id),
                           stringsAsFactors = FALSE)
    corrgene$unspliced <- paste0(corrgene$spliced, suffixes["unspliced"])
    corrgene$intron <- paste0(corrgene$spliced, suffixes["intron"])

    ## --------------------------------------------------------------------- ##
    ## Spliced transcripts
    ## --------------------------------------------------------------------- ##
    if ("spliced" %in% featureType) {
        if (verbose) {
            message("Extracting spliced transcript features")
        }

        featurelist$spliced <- names(ebt)

        ## Here, it's important that for each transcript, the exons must be
        ## ordered by ascending rank, that is, by ascending position in
        ## the transcript.
        grlfull <- c(grlfull, ebt)
    }

    ## --------------------------------------------------------------------- ##
    ## Unspliced transcripts
    ## --------------------------------------------------------------------- ##
    if ("unspliced" %in% featureType) {
        if (verbose) {
            message("Extracting unspliced transcript features")
        }

        ebtr <- range(ebt)
        e2 <- BiocGenerics::unlist(ebtr)
        e2$exon_rank <- 1L
        e2$transcript_id <- names(e2)
        e2$gene_id <- t2g$GENEID[match(e2$transcript_id, t2g$TXNAME)]
        e2$type <- "exon"
        e2$transcript_id <- paste0(e2$transcript_id, suffixes["unspliced"])
        e2$gene_id <- paste0(e2$gene_id, suffixes["unspliced"])
        e2$exon_id <- e2$transcript_id
        names(e2) <- NULL
        mcols(e2) <-
            S4Vectors::mcols(e2)[, c("exon_id", "exon_rank",
                                     "transcript_id", "gene_id", "type")]
        ebtr <- BiocGenerics::relist(e2, ebtr)
        names(ebtr) <- paste0(names(ebtr), suffixes["unspliced"])

        featurelist$unspliced <- names(ebtr)
        grlfull <- c(grlfull, ebtr)
    }

    ## --------------------------------------------------------------------- ##
    ## Introns
    ## --------------------------------------------------------------------- ##
    if ("intron" %in% featureType) {
        if (verbose) {
            message("Extracting introns using the ", intronType, " approach")
        }

        if (intronType == "separate") {
            ## Group exons by transcript
            grl <- ebt
        } else if (intronType == "collapse") {
            ## Group exons by gene
            grl <- ebg

            ## Collapse the exons of each gene
            grl <- GenomicRanges::reduce(grl)
        }

        ## Get the full range of each feature - will be used to check
        ## that the introns don't go outside after adding the flank
        grlrange <- range(grl)

        ## Get introns as the set difference between the range and the exons,
        ## for each transcript
        ## Here, the order of the introns doesn't really matter, since they
        ## will not be joined together into a transcript
        grl <- BiocGenerics::setdiff(grlrange, grl)

        ## Remove empty entries
        keep <- lengths(grl) > 0L
        grl <- grl[keep]
        grlrange <- grlrange[keep]

        ## Add flanking region
        grl <- grl + flankLength

        ## Make sure features (including flanks) are within the
        ## boundaries of the corresponding feature
        if (keepIntronInFeature) {
            grl <- GenomicRanges::restrict(
                grl,
                start = BiocGenerics::start(unlist(grlrange)),
                end = BiocGenerics::end(unlist(grlrange))
            )
        }

        if (joinOverlappingIntrons) {
            ## If two (introns + flankLength) overlap, join them
            grl <- GenomicRanges::reduce(grl)
        }

        gr <- BiocGenerics::unlist(grl)
        gr$exon_rank <- 1L
        gr$transcript_id <- names(gr)
        if (intronType == "separate") {
            gr$gene_id <- t2g$GENEID[match(gr$transcript_id, t2g$TXNAME)]
        } else {
            gr$gene_id <- gr$transcript_id
        }

        if (collapseIntronsByGene) {
            ## Split introns by gene and join any overlapping introns
            grl <- S4Vectors::split(gr, f = gr$gene_id)
            grl <- GenomicRanges::reduce(grl)

            ## Unlist reduced introns
            gr <- BiocGenerics::unlist(grl)

            gr$exon_rank <- 1L

            ## After collapsing, the names will be gene IDs
            gr$transcript_id <- names(gr)
            gr$gene_id <- gr$transcript_id
        }

        gr$type <- "exon"
        gr$transcript_id <- gsub(
            paste0(suffixes["intron"], "."), suffixes["intron"],
            make.unique(paste0(gr$transcript_id, suffixes["intron"])),
            fixed = TRUE
        )
        gr$gene_id <- paste0(gr$gene_id, suffixes["intron"])
        gr$exon_id <- gr$transcript_id
        names(gr) <- NULL
        mcols(gr) <-
            S4Vectors::mcols(gr)[, c("exon_id", "exon_rank",
                                     "transcript_id", "gene_id", "type")]
        grl <- BiocGenerics::relist(gr, lapply(
            setNames(seq_along(gr), gr$transcript_id), function(i) i
        ))

        featurelist$intron <- names(grl)

        grlfull <- c(grlfull, grl)
    }

    S4Vectors::metadata(grlfull)$corrtx <-
        corrtx[, colnames(corrtx) %in% featureType, drop = FALSE]
    S4Vectors::metadata(grlfull)$corrgene <-
        corrgene[, colnames(corrgene) %in% featureType, drop = FALSE]
    S4Vectors::metadata(grlfull)$featurelist <- featurelist

    grlfull
}

context("Reference generation")

library(BSgenome)

gtf <- system.file("extdata", "small_example.gtf", package = "eisaR")
txome <- system.file("extdata", "small_example_txome.fa", package = "eisaR")
genome <- system.file("extdata", "small_example_genome.fa", package = "eisaR")
t2g <- system.file("extdata", "small_example_t2g.txt", package = "eisaR")

test_that("reference extraction fails with the wrong inputs", {
    expect_error(getFeatureRanges(gtf = "nonexistent"),
                 regexp = "'gtf' must be a character scalar")
    expect_error(getFeatureRanges(gtf = 1L),
                 regexp = "'gtf' must be a character scalar")
    expect_error(getFeatureRanges(gtf = c(gtf, gtf)),
                 regexp = "'gtf' must be a character scalar")

    expect_error(getFeatureRanges(gtf = gtf, featureType = "wrongvalue"),
                 regexp = "'featureType' must be a subset of")
    expect_error(getFeatureRanges(gtf = gtf, featureType = 1L),
                 regexp = "'featureType' must be a subset of")

    ## Check intron-related arguments if 'intron' is in featureType
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "wrongvalue"),
                 regexp = "'intronType' must be either")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = 1.0),
                 regexp = "'intronType' must be either")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = c("separate", "collapse")),
                 regexp = "'intronType' must be either")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = "string"),
                 regexp = "'flankLength' must be a numeric")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = -1L),
                 regexp = "'flankLength' must be a numeric")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = c(1L, 2L)),
                 regexp = "'flankLength' must be a numeric")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = "TRUE"),
                 regexp = "'joinOverlappingIntrons' must be a logical")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = c(TRUE, FALSE)),
                 regexp = "'joinOverlappingIntrons' must be a logical")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = FALSE,
                                  collapseIntronsByGene = "TRUE"),
                 regexp = "'collapseIntronsByGene' must be a logical")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = FALSE,
                                  collapseIntronsByGene = c(TRUE, FALSE)),
                 regexp = "'collapseIntronsByGene' must be a logical")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = FALSE,
                                  keepIntronInFeature = "TRUE"),
                 regexp = "'keepIntronInFeature' must be a logical")
    expect_error(getFeatureRanges(gtf = gtf, featureType = c("spliced", "intron"),
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = FALSE,
                                  keepIntronInFeature = c(TRUE, FALSE)),
                 regexp = "'keepIntronInFeature' must be a logical")

    ## If 'intron' is not in featureType, intron-related arguments are not checked
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "wrongvalue"),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = 1.0),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = c("separate", "collapse")),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = "string"),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = -1L),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = c(1L, 2L)),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = "TRUE"),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = 2L,
                                  joinOverlappingIntrons = c(TRUE, FALSE)),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = 2L,
                                  collapseIntronsByGene = "TRUE"),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = 2L,
                                  collapseIntronsByGene = c(TRUE, FALSE)),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = 2L,
                                  keepIntronInFeature = "TRUE"),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, featureType = "spliced", verbose = "TRUE",
                                  intronType = "separate", flankLength = 2L,
                                  keepIntronInFeature = c(TRUE, FALSE)),
                 regexp = "'verbose' must be a logical scalar")

    expect_error(getFeatureRanges(gtf = gtf, verbose = 1L),
                 regexp = "'verbose' must be a logical scalar")
    expect_error(getFeatureRanges(gtf = gtf, verbose = c(TRUE, FALSE)),
                 regexp = "'verbose' must be a logical scalar")
})

test_that("feature extraction fails if there are missing required packages", {
    # these tests assume that all non-base packages are installed in
    # .Library.site and will fail if this is not the case (e.g. on BioC builders)
    skip_on_bioc()
    # also skip on r-universe (https://docs.r-universe.dev/publish/troubleshoot-build.html#how-to-know-whether-tests-are-run-on-r-universe)
    skip_if_not(identical(Sys.getenv("MY_UNIVERSE"), ""))

    # set new lib paths
    # ... unload ns
    unloadNamespace("ensembldb")
    unloadNamespace("txdbmaker")
    unloadNamespace("GenomicFeatures")
    withr::with_temp_libpaths({
        expect_error(getFeatureRanges(gtf = gtf))
    }, action = "replace")
    # ... reload ns
    loadNamespace("GenomicFeatures")
    loadNamespace("txdbmaker")
    loadNamespace("ensembldb")

})

test_that("feature range extraction works", {
    tx2gene <- read.delim(t2g, header = FALSE, as.is = TRUE)

    ## ----------------------------------------------------------------------- ##
    ## 'Separate' intron definition
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "separate",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    expect_length(frs, 25L)

    expect_identical(unique(unlist(frs)$type), "exon")
    expect_named(unlist(frs), unlist(frs)$transcript_id)

    ## Spliced transcripts
    splicedIds <- metadata(frs)$corrtx$spliced
    expect_identical(sort(splicedIds), sort(tx2gene$V1))
    txgtf <- subset(rtracklayer::import(gtf), type == "exon")
    txgtfl <- as(split(txgtf, f = txgtf$transcript_id), "GRangesList")
    txfrs <- frs[names(txgtfl)]
    expect_named(txfrs, splicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrs), ranges(txgtfl))

    expect_identical(unique(frs$tx1.1$gene_id), "g1")
    expect_identical(unique(frs$tx1.2$gene_id), "g1")
    expect_identical(unique(frs$tx1.3$gene_id), "g1")
    expect_identical(unique(frs$`tx2.1-I`$gene_id), "g2-I")
    expect_identical(unique(frs$`tx2.2-U`$gene_id), "g2-I")
    expect_identical(unique(frs$tx3.1$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.2$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.3$gene_id), "g3-U")

    ## Unspliced transcripts
    unsplicedIds <- metadata(frs)$corrtx$unspliced
    txgtflr <- range(txgtfl)
    names(txgtflr) <- paste0(names(txgtflr), "-U")
    txfrsr <- frs[names(txgtflr)]
    expect_named(txfrsr, unsplicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrsr), ranges(txgtflr))

    expect_identical(unique(frs$`tx1.1-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.2-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.3-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx2.1-I-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx2.2-U-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx3.1-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.2-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.3-U`$gene_id), "g3-U-U")

    ## Introns
    expect_identical(ranges(frs$`tx1.1-I`), IRanges(start = 118L, end = 143L))
    expect_identical(as.character(seqnames(frs$`tx1.1-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.1-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I`), IRanges(start = 78L, end = 103L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I1`), IRanges(start = 118L, end = 143L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I1`)), "+")
    expect_identical(ranges(frs$`tx3.1-I`), IRanges(start = 88L, end = 143L))
    expect_identical(as.character(seqnames(frs$`tx3.1-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.1-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I`), IRanges(start = 78L, end = 103L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I1`), IRanges(start = 118L, end = 143L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I1`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I1`)), "+")
    expect_identical(ranges(frs$`tx2.1-I-I`), IRanges(start = 28L, end = 53L))
    expect_identical(as.character(seqnames(frs$`tx2.1-I-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.1-I-I`)), "-")
    expect_identical(ranges(frs$`tx2.2-U-I`), IRanges(start = 18L, end = 38L))
    expect_identical(as.character(seqnames(frs$`tx2.2-U-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.2-U-I`)), "-")
    expect_identical(ranges(frs$`tx2.2-U-I1`), IRanges(start = 43L, end = 63L))
    expect_identical(as.character(seqnames(frs$`tx2.2-U-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.2-U-I1`)), "-")

    expect_identical(unique(frs$`tx1.1-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx1.2-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx1.2-I1`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx2.1-I-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx2.2-U-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx2.2-U-I1`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx3.1-I`$gene_id), "g3-U-I")
    expect_identical(unique(frs$`tx3.2-I`$gene_id), "g3-U-I")
    expect_identical(unique(frs$`tx3.2-I1`$gene_id), "g3-U-I")


    ## ----------------------------------------------------------------------- ##
    ## 'Separate' intron definition, join overlapping introns
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "separate",
                            flankLength = 9L,
                            joinOverlappingIntrons = TRUE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    expect_length(frs, 24L)

    expect_identical(unique(unlist(frs)$type), "exon")
    expect_named(unlist(frs), unlist(frs)$transcript_id)

    ## Spliced transcripts
    splicedIds <- metadata(frs)$corrtx$spliced
    expect_identical(sort(splicedIds), sort(tx2gene$V1))
    txgtf <- subset(rtracklayer::import(gtf), type == "exon")
    txgtfl <- as(split(txgtf, f = txgtf$transcript_id), "GRangesList")
    txfrs <- frs[names(txgtfl)]
    expect_named(txfrs, splicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrs), ranges(txgtfl))

    ## Unspliced transcripts
    unsplicedIds <- metadata(frs)$corrtx$unspliced
    txgtflr <- range(txgtfl)
    names(txgtflr) <- paste0(names(txgtflr), "-U")
    txfrsr <- frs[names(txgtflr)]
    expect_named(txfrsr, unsplicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrsr), ranges(txgtflr))

    ## Introns
    expect_identical(ranges(frs$`tx1.1-I`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx1.1-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.1-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I`), IRanges(start = 72L, end = 109L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I1`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I1`)), "+")
    expect_identical(ranges(frs$`tx3.1-I`), IRanges(start = 82L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx3.1-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.1-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I`), IRanges(start = 72L, end = 109L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I1`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I1`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I1`)), "+")
    expect_identical(ranges(frs$`tx2.1-I-I`), IRanges(start = 22L, end = 59L))
    expect_identical(as.character(seqnames(frs$`tx2.1-I-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.1-I-I`)), "-")
    expect_identical(ranges(frs$`tx2.2-U-I`), IRanges(start = 12L, end = 69L))
    expect_identical(as.character(seqnames(frs$`tx2.2-U-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.2-U-I`)), "-")

    ## ----------------------------------------------------------------------- ##
    ## Collapse introns by gene
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "separate",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = TRUE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    expect_length(frs, 20L)

    expect_identical(unique(unlist(frs)$type), "exon")
    expect_named(unlist(frs), unlist(frs)$transcript_id)

    ## Spliced transcripts
    splicedIds <- metadata(frs)$corrtx$spliced
    expect_identical(sort(splicedIds), sort(tx2gene$V1))
    txgtf <- subset(rtracklayer::import(gtf), type == "exon")
    txgtfl <- as(split(txgtf, f = txgtf$transcript_id), "GRangesList")
    txfrs <- frs[names(txgtfl)]
    expect_named(txfrs, splicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrs), ranges(txgtfl))

    expect_identical(unique(frs$tx1.1$gene_id), "g1")
    expect_identical(unique(frs$tx1.2$gene_id), "g1")
    expect_identical(unique(frs$tx1.3$gene_id), "g1")
    expect_identical(unique(frs$`tx2.1-I`$gene_id), "g2-I")
    expect_identical(unique(frs$`tx2.2-U`$gene_id), "g2-I")
    expect_identical(unique(frs$tx3.1$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.2$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.3$gene_id), "g3-U")

    ## Unspliced transcripts
    unsplicedIds <- metadata(frs)$corrtx$unspliced
    txgtflr <- range(txgtfl)
    names(txgtflr) <- paste0(names(txgtflr), "-U")
    txfrsr <- frs[names(txgtflr)]
    expect_named(txfrsr, unsplicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrsr), ranges(txgtflr))

    expect_identical(unique(frs$`tx1.1-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.2-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.3-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx2.1-I-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx2.2-U-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx3.1-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.2-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.3-U`$gene_id), "g3-U-U")

    ## Introns
    expect_identical(ranges(frs$`g1-I`), IRanges(start = 78L, end = 103L))
    expect_identical(as.character(seqnames(frs$`g1-I`)), "chr1")
    expect_identical(as.character(strand(frs$`g1-I`)), "+")
    expect_identical(ranges(frs$`g1-I1`), IRanges(start = 118L, end = 143L))
    expect_identical(as.character(seqnames(frs$`g1-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`g1-I1`)), "+")
    expect_identical(ranges(frs$`g3-U-I`), IRanges(start = 78L, end = 143L))
    expect_identical(as.character(seqnames(frs$`g3-U-I`)), "chr2")
    expect_identical(as.character(strand(frs$`g3-U-I`)), "+")
    expect_identical(ranges(frs$`g2-I-I`), IRanges(start = 18L, end = 63L))
    expect_identical(as.character(seqnames(frs$`g2-I-I`)), "chr1")
    expect_identical(as.character(strand(frs$`g2-I-I`)), "-")

    expect_identical(unique(frs$`g1-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`g1-I1`$gene_id), "g1-I")
    expect_identical(unique(frs$`g2-I-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`g3-U-I`$gene_id), "g3-U-I")

    ## ----------------------------------------------------------------------- ##
    ## Introns extending outside of feature
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "separate",
                            flankLength = 9L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    expect_length(frs, 25L)

    expect_identical(unique(unlist(frs)$type), "exon")
    expect_named(unlist(frs), unlist(frs)$transcript_id)

    ## Spliced transcripts
    splicedIds <- metadata(frs)$corrtx$spliced
    expect_identical(sort(splicedIds), sort(tx2gene$V1))
    txgtf <- subset(rtracklayer::import(gtf), type == "exon")
    txgtfl <- as(split(txgtf, f = txgtf$transcript_id), "GRangesList")
    txfrs <- frs[names(txgtfl)]
    expect_named(txfrs, splicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrs), ranges(txgtfl))

    expect_identical(unique(frs$tx1.1$gene_id), "g1")
    expect_identical(unique(frs$tx1.2$gene_id), "g1")
    expect_identical(unique(frs$tx1.3$gene_id), "g1")
    expect_identical(unique(frs$`tx2.1-I`$gene_id), "g2-I")
    expect_identical(unique(frs$`tx2.2-U`$gene_id), "g2-I")
    expect_identical(unique(frs$tx3.1$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.2$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.3$gene_id), "g3-U")

    ## Unspliced transcripts
    unsplicedIds <- metadata(frs)$corrtx$unspliced
    txgtflr <- range(txgtfl)
    names(txgtflr) <- paste0(names(txgtflr), "-U")
    txfrsr <- frs[names(txgtflr)]
    expect_named(txfrsr, unsplicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrsr), ranges(txgtflr))

    expect_identical(unique(frs$`tx1.1-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.2-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.3-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx2.1-I-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx2.2-U-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx3.1-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.2-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.3-U`$gene_id), "g3-U-U")

    ## Introns
    expect_identical(ranges(frs$`tx1.1-I`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx1.1-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.1-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I`), IRanges(start = 72L, end = 109L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I1`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I1`)), "+")
    expect_identical(ranges(frs$`tx3.1-I`), IRanges(start = 82L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx3.1-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.1-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I`), IRanges(start = 72L, end = 109L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I1`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I1`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I1`)), "+")
    expect_identical(ranges(frs$`tx2.1-I-I`), IRanges(start = 22L, end = 59L))
    expect_identical(as.character(seqnames(frs$`tx2.1-I-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.1-I-I`)), "-")
    expect_identical(ranges(frs$`tx2.2-U-I`), IRanges(start = 12L, end = 44L))
    expect_identical(as.character(seqnames(frs$`tx2.2-U-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.2-U-I`)), "-")
    expect_identical(ranges(frs$`tx2.2-U-I1`), IRanges(start = 37L, end = 69L))
    expect_identical(as.character(seqnames(frs$`tx2.2-U-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.2-U-I1`)), "-")

    expect_identical(unique(frs$`tx1.1-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx1.2-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx1.2-I1`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx2.1-I-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx2.2-U-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx2.2-U-I1`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx3.1-I`$gene_id), "g3-U-I")
    expect_identical(unique(frs$`tx3.2-I`$gene_id), "g3-U-I")
    expect_identical(unique(frs$`tx3.2-I1`$gene_id), "g3-U-I")


    ## ----------------------------------------------------------------------- ##
    ## Keep introns within feature
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "separate",
                            flankLength = 9L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = TRUE,
                            verbose = FALSE)
    expect_length(frs, 25L)

    expect_identical(unique(unlist(frs)$type), "exon")
    expect_named(unlist(frs), unlist(frs)$transcript_id)

    ## Spliced transcripts
    splicedIds <- metadata(frs)$corrtx$spliced
    expect_identical(sort(splicedIds), sort(tx2gene$V1))
    txgtf <- subset(rtracklayer::import(gtf), type == "exon")
    txgtfl <- as(split(txgtf, f = txgtf$transcript_id), "GRangesList")
    txfrs <- frs[names(txgtfl)]
    expect_named(txfrs, splicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrs), ranges(txgtfl))

    expect_identical(unique(frs$tx1.1$gene_id), "g1")
    expect_identical(unique(frs$tx1.2$gene_id), "g1")
    expect_identical(unique(frs$tx1.3$gene_id), "g1")
    expect_identical(unique(frs$`tx2.1-I`$gene_id), "g2-I")
    expect_identical(unique(frs$`tx2.2-U`$gene_id), "g2-I")
    expect_identical(unique(frs$tx3.1$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.2$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.3$gene_id), "g3-U")

    ## Unspliced transcripts
    unsplicedIds <- metadata(frs)$corrtx$unspliced
    txgtflr <- range(txgtfl)
    names(txgtflr) <- paste0(names(txgtflr), "-U")
    txfrsr <- frs[names(txgtflr)]
    expect_named(txfrsr, unsplicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrsr), ranges(txgtflr))

    expect_identical(unique(frs$`tx1.1-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.2-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.3-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx2.1-I-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx2.2-U-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx3.1-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.2-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.3-U`$gene_id), "g3-U-U")

    ## Introns
    expect_identical(ranges(frs$`tx1.1-I`), IRanges(start = 116L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx1.1-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.1-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I`), IRanges(start = 72L, end = 109L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I`)), "+")
    expect_identical(ranges(frs$`tx1.2-I1`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx1.2-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`tx1.2-I1`)), "+")
    expect_identical(ranges(frs$`tx3.1-I`), IRanges(start = 82L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx3.1-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.1-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I`), IRanges(start = 72L, end = 109L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I`)), "+")
    expect_identical(ranges(frs$`tx3.2-I1`), IRanges(start = 112L, end = 149L))
    expect_identical(as.character(seqnames(frs$`tx3.2-I1`)), "chr2")
    expect_identical(as.character(strand(frs$`tx3.2-I1`)), "+")
    expect_identical(ranges(frs$`tx2.1-I-I`), IRanges(start = 22L, end = 59L))
    expect_identical(as.character(seqnames(frs$`tx2.1-I-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.1-I-I`)), "-")
    expect_identical(ranges(frs$`tx2.2-U-I`), IRanges(start = 12L, end = 44L))
    expect_identical(as.character(seqnames(frs$`tx2.2-U-I`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.2-U-I`)), "-")
    expect_identical(ranges(frs$`tx2.2-U-I1`), IRanges(start = 37L, end = 69L))
    expect_identical(as.character(seqnames(frs$`tx2.2-U-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`tx2.2-U-I1`)), "-")

    expect_identical(unique(frs$`tx1.1-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx1.2-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx1.2-I1`$gene_id), "g1-I")
    expect_identical(unique(frs$`tx2.1-I-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx2.2-U-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx2.2-U-I1`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`tx3.1-I`$gene_id), "g3-U-I")
    expect_identical(unique(frs$`tx3.2-I`$gene_id), "g3-U-I")
    expect_identical(unique(frs$`tx3.2-I1`$gene_id), "g3-U-I")


    ## ----------------------------------------------------------------------- ##
    ## 'Collapse' intron definition
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "collapse",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    expect_length(frs, 22L)

    expect_identical(unique(unlist(frs)$type), "exon")
    expect_named(unlist(frs), unlist(frs)$transcript_id)

    ## Spliced transcripts
    splicedIds <- metadata(frs)$corrtx$spliced
    expect_identical(sort(splicedIds), sort(tx2gene$V1))
    txgtf <- subset(rtracklayer::import(gtf), type == "exon")
    txgtfl <- as(split(txgtf, f = txgtf$transcript_id), "GRangesList")
    txfrs <- frs[names(txgtfl)]
    expect_named(txfrs, splicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrs), ranges(txgtfl))

    expect_identical(unique(frs$tx1.1$gene_id), "g1")
    expect_identical(unique(frs$tx1.2$gene_id), "g1")
    expect_identical(unique(frs$tx1.3$gene_id), "g1")
    expect_identical(unique(frs$`tx2.1-I`$gene_id), "g2-I")
    expect_identical(unique(frs$`tx2.2-U`$gene_id), "g2-I")
    expect_identical(unique(frs$tx3.1$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.2$gene_id), "g3-U")
    expect_identical(unique(frs$tx3.3$gene_id), "g3-U")

    ## Unspliced transcripts
    unsplicedIds <- metadata(frs)$corrtx$unspliced
    txgtflr <- range(txgtfl)
    names(txgtflr) <- paste0(names(txgtflr), "-U")
    txfrsr <- frs[names(txgtflr)]
    expect_named(txfrsr, unsplicedIds, ignore.order = TRUE)
    expect_identical(ranges(txfrsr), ranges(txgtflr))

    expect_identical(unique(frs$`tx1.1-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.2-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx1.3-U`$gene_id), "g1-U")
    expect_identical(unique(frs$`tx2.1-I-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx2.2-U-U`$gene_id), "g2-I-U")
    expect_identical(unique(frs$`tx3.1-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.2-U`$gene_id), "g3-U-U")
    expect_identical(unique(frs$`tx3.3-U`$gene_id), "g3-U-U")

    ## Introns
    expect_identical(ranges(frs$`g1-I`), IRanges(start = 88L, end = 103L))
    expect_identical(as.character(seqnames(frs$`g1-I`)), "chr1")
    expect_identical(as.character(strand(frs$`g1-I`)), "+")
    expect_identical(ranges(frs$`g1-I1`), IRanges(start = 118L, end = 143L))
    expect_identical(as.character(seqnames(frs$`g1-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`g1-I1`)), "+")
    expect_identical(ranges(frs$`g2-I-I`), IRanges(start = 28L, end = 38L))
    expect_identical(as.character(seqnames(frs$`g2-I-I`)), "chr1")
    expect_identical(as.character(strand(frs$`g2-I-I`)), "-")
    expect_identical(ranges(frs$`g2-I-I1`), IRanges(start = 43L, end = 53L))
    expect_identical(as.character(seqnames(frs$`g2-I-I1`)), "chr1")
    expect_identical(as.character(strand(frs$`g2-I-I1`)), "-")
    expect_identical(ranges(frs$`g3-U-I`), IRanges(start = 88L, end = 103L))
    expect_identical(as.character(seqnames(frs$`g3-U-I`)), "chr2")
    expect_identical(as.character(strand(frs$`g3-U-I`)), "+")
    expect_identical(ranges(frs$`g3-U-I1`), IRanges(start = 118L, end = 143L))
    expect_identical(as.character(seqnames(frs$`g3-U-I1`)), "chr2")
    expect_identical(as.character(strand(frs$`g3-U-I1`)), "+")

    expect_identical(unique(frs$`g1-I`$gene_id), "g1-I")
    expect_identical(unique(frs$`g1-I1`$gene_id), "g1-I")
    expect_identical(unique(frs$`g2-I-I`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`g2-I-I1`$gene_id), "g2-I-I")
    expect_identical(unique(frs$`g3-U-I`$gene_id), "g3-U-I")
    expect_identical(unique(frs$`g3-U-I1`$gene_id), "g3-U-I")

})

test_that("feature sequence extraction works", {
    ## ----------------------------------------------------------------------- ##
    ## 'Separate' intron definition
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "separate",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    expect_length(frs, 25L)
    expect_length(metadata(frs)$featurelist, 3L)
    expect_length(metadata(frs)$featurelist$spliced, 8L)
    expect_length(metadata(frs)$featurelist$unspliced, 8L)
    expect_length(metadata(frs)$featurelist$intron, 9L)

    genomeseq <- Biostrings::readDNAStringSet(genome)
    seqs <- GenomicFeatures::extractTranscriptSeqs(x = genomeseq, transcripts = frs)

    txs <- Biostrings::readDNAStringSet(txome)

    ## Spliced transcripts
    txseqs <- seqs[names(txs)]
    expect_equivalent(txs, txseqs)
    expect_identical(as.character(txs), as.character(txseqs))

    ## Unspliced transcripts
    expect_identical(as.character(seqs[["tx1.1-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 116L, 160L)))
    expect_identical(as.character(seqs[["tx1.2-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 61L, 160L)))
    expect_identical(as.character(seqs[["tx1.3-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 61L, 90L)))
    expect_identical(as.character(seqs[["tx2.1-I-U"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 11L, 70L)
                     )))
    expect_identical(as.character(seqs[["tx2.2-U-U"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 11L, 80L)
                     )))
    expect_identical(as.character(seqs[["tx3.1-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 61L, 160L)))
    expect_identical(as.character(seqs[["tx3.2-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 61L, 160L)))
    expect_identical(as.character(seqs[["tx3.3-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 61L, 90L)))

    ## Introns
    expect_identical(as.character(seqs[["tx1.1-I"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 118L, 143L)))
    expect_identical(as.character(seqs[["tx1.2-I"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 78L, 103L)))
    expect_identical(as.character(seqs[["tx1.2-I1"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 118L, 143L)))
    expect_identical(as.character(seqs[["tx2.1-I-I"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 28L, 53L)
                     )))
    expect_identical(as.character(seqs[["tx2.2-U-I"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 18L, 38L)
                     )))
    expect_identical(as.character(seqs[["tx2.2-U-I1"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 43L, 63L)
                     )))
    expect_identical(as.character(seqs[["tx3.1-I"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 88L, 143L)))
    expect_identical(as.character(seqs[["tx3.2-I"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 78L, 103L)))
    expect_identical(as.character(seqs[["tx3.2-I1"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 118L, 143L)))

    ## ----------------------------------------------------------------------- ##
    ## 'Collapse' intron definition
    ## ----------------------------------------------------------------------- ##
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "unspliced", "intron"),
                            intronType = "collapse",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    expect_length(frs, 22L)
    expect_length(metadata(frs)$featurelist, 3L)
    expect_length(metadata(frs)$featurelist$spliced, 8L)
    expect_length(metadata(frs)$featurelist$unspliced, 8L)
    expect_length(metadata(frs)$featurelist$intron, 6L)

    genomeseq <- Biostrings::readDNAStringSet(genome)
    seqs <- GenomicFeatures::extractTranscriptSeqs(x = genomeseq, transcripts = frs)

    txs <- Biostrings::readDNAStringSet(txome)

    ## Spliced transcripts
    txseqs <- seqs[names(txs)]
    expect_equivalent(txs, txseqs)
    expect_identical(as.character(txs), as.character(txseqs))

    ## Unspliced transcripts
    expect_identical(as.character(seqs[["tx1.1-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 116L, 160L)))
    expect_identical(as.character(seqs[["tx1.2-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 61L, 160L)))
    expect_identical(as.character(seqs[["tx1.3-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 61L, 90L)))
    expect_identical(as.character(seqs[["tx2.1-I-U"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 11L, 70L)
                     )))
    expect_identical(as.character(seqs[["tx2.2-U-U"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 11L, 80L)
                     )))
    expect_identical(as.character(seqs[["tx3.1-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 61L, 160L)))
    expect_identical(as.character(seqs[["tx3.2-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 61L, 160L)))
    expect_identical(as.character(seqs[["tx3.3-U"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 61L, 90L)))

    ## Introns
    expect_identical(as.character(seqs[["g1-I"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 88L, 103L)))
    expect_identical(as.character(seqs[["g1-I1"]]),
                     as.character(Biostrings::substr(genomeseq[["chr1"]], 118L, 143L)))
    expect_identical(as.character(seqs[["g2-I-I"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 28L, 38L)
                     )))
    expect_identical(as.character(seqs[["g2-I-I1"]]),
                     as.character(Biostrings::reverseComplement(
                         Biostrings::substr(genomeseq[["chr1"]], 43L, 53L)
                     )))
    expect_identical(as.character(seqs[["g3-U-I"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 88L, 103L)))
    expect_identical(as.character(seqs[["g3-U-I1"]]),
                     as.character(Biostrings::substr(genomeseq[["chr2"]], 118L, 143L)))
})

test_that("gtf export works", {
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "intron"),
                            intronType = "collapse",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)
    # arguments
    f <- file.path(tempdir(), "exported_gtf.gtf")
    expect_error(exportToGtf("wrongtype", filepath = f),
                 regexp = "'grl' must be a GRangesList")
    exportToGtf(frs, filepath = f)
    rb <- rtracklayer::import(f)

    # expected results
    expect_length(rb, 17L + 6L + 8L + 6L + 3L + 3L)
    expect_identical(ranges(subset(rb, gene_id == "g1" & type == "gene")),
                     IRanges(start = 61L, end = 160L))
    expect_identical(as.character(seqnames(subset(rb, gene_id == "g1" & type == "gene"))),
                     "chr1")
    expect_identical(as.character(strand(subset(rb, gene_id == "g1" & type == "gene"))),
                     "+")

    expect_identical(ranges(subset(rb, transcript_id == "tx1.3" & type == "transcript")),
                     IRanges(start = 61L, end = 90L))
    expect_identical(
        as.character(
            seqnames(subset(rb, transcript_id == "tx1.3" & type == "transcript"))
        ),
        "chr1"
    )
    expect_identical(
        as.character(
            strand(subset(rb, transcript_id == "tx1.3" & type == "transcript"))
        ),
        "+"
    )
})

test_that("gtf export fails if required packages are missing", {
    # these tests assume that all non-base packages are installed in
    # .Library.site and will fail if this is not the case (e.g. on BioC builders)
    skip_on_bioc()
    # also skip on r-universe (https://docs.r-universe.dev/publish/troubleshoot-build.html#how-to-know-whether-tests-are-run-on-r-universe)
    skip_if_not(identical(Sys.getenv("MY_UNIVERSE"), ""))

    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "intron"),
                            intronType = "collapse",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = FALSE)

    # set new lib paths
    # ... unload ns
    unloadNamespace("ensembldb")
    unloadNamespace("txdbmaker")
    unloadNamespace("GenomicFeatures")
    unloadNamespace("BSgenome")
    unloadNamespace("rtracklayer")
    withr::with_temp_libpaths({
        expect_error(exportToGtf(frs, tempfile()))
    }, action = "replace")
    # ... load ns
    loadNamespace("rtracklayer")
    loadNamespace("GenomicFeatures")
    loadNamespace("BSgenome")
    loadNamespace("txdbmaker")
    loadNamespace("ensembldb")
})

test_that("tx2gene generation works", {
    frs <- getFeatureRanges(gtf = gtf,
                            featureType = c("spliced", "intron", "unspliced"),
                            intronType = "collapse",
                            flankLength = 3L,
                            joinOverlappingIntrons = FALSE,
                            collapseIntronsByGene = FALSE,
                            keepIntronInFeature = FALSE,
                            verbose = TRUE)
    getTx2Gene(frs, file.path(tempdir(), "tx2gene.tsv"))
    txg <- getTx2Gene(frs)

    ## Generate expected tx2gene mapping
    df <- read.delim(t2g, header = FALSE, as.is = TRUE)
    colnames(df) <- c("transcript_id", "gene_id")
    dfu <- df
    dfu$transcript_id <- paste0(dfu$transcript_id, "-U")
    dfu$gene_id <- paste0(dfu$gene_id, "-U")
    df <- rbind(df, dfu)
    df <- rbind(df, data.frame(transcript_id = c("g1-I", "g1-I1", "g2-I-I",
                                                 "g2-I-I1", "g3-U-I", "g3-U-I1"),
                               gene_id = c("g1-I", "g1-I", "g2-I-I", "g2-I-I",
                                           "g3-U-I", "g3-U-I")))
    df <- df[order(df$transcript_id, df$gene_id), ]
    rownames(df) <- NULL

    txg <- txg[order(txg$transcript_id, txg$gene_id), ]
    rownames(txg) <- NULL

    expect_identical(df, txg)
})

#' Export GRangesList to GTF
#'
#' Export the features in a GRangesList generated by \code{getFeatureRanges}
#' to a GTF file. The function will represent each row of each of the
#' entries as an "exon", each individual entry as a "transcript", and
#' aggregate all features belonging to the same gene as a "gene" entry in
#' the GTF file.
#'
#' @param grl \code{GRangesList} object, typically generated by
#'   \code{getFeatureRanges}
#' @param filepath Path to output GTF file
#'
#' @author Charlotte Soneson
#'
#' @export
#'
#' @return Does not return anything, generates a GTF file
#'
#' @importFrom BiocGenerics sort unlist
#' @importFrom S4Vectors split
#'
#' @examples
#'   ## Get feature ranges
#'   grl <- getFeatureRanges(
#'     gtf = system.file("extdata/small_example.gtf", package = "eisaR"),
#'     featureType = c("spliced", "intron"),
#'     intronType = "separate",
#'     flankLength = 5L,
#'     joinOverlappingIntrons = FALSE,
#'     verbose = TRUE
#'   )
#'
#'   ## Export GTF
#'   exportToGtf(grl = grl, filepath = file.path(tempdir(), "exported.gtf"))
#'
exportToGtf <- function(grl, filepath) {
    if (!is(grl, "GRangesList")) {
        stop("'grl' must be a GRangesList")
    }

    if (!requireNamespace("rtracklayer", quietly = TRUE)) {
        stop("exportToGtf() requires installing the Bioconductor package 'rtracklayer'",
             " using BiocManager::install(\"rtracklayer\")")
    }

    ## "Exons" (individual exons for spliced transcripts, unspliced transcripts
    ## and introns separately)
    gre <- BiocGenerics::unlist(grl)

    ## "Transcripts" (full transcript range for spliced transcripts, unspliced
    ## transcripts and introns separately)
    grt <- BiocGenerics::unlist(range(grl))
    grt$exon_id <- NA
    grt$exon_rank <- NA
    grt$transcript_id <- names(grt)
    grt$gene_id <- gre$gene_id[match(grt$transcript_id, gre$transcript_id)]
    grt$type <- "transcript"

    ## "Gene"
    grg <- S4Vectors::split(gre, gre$gene_id)
    grg <- BiocGenerics::unlist(range(grg))
    grg$exon_id <- NA
    grg$exon_rank <- NA
    grg$transcript_id <- NA
    grg$gene_id <- names(grg)
    grg$type <- "gene"

    grtot <- BiocGenerics::sort(c(gre, grt, grg))
    rtracklayer::export(grtot, con = filepath, format = "gtf")
}

#' @title Visualize the results from an exon-intron split analysis.
#'
#' @description \code{plotEISA} takes the return value from \code{\link{runEISA}}
#'   and generates a scatterplot of intronic versus exonic changes.
#'
#' @author Michael Stadler
#'
#' @param x \code{list} with EISA results, typically the return value from \code{\link{runEISA}}
#' @param contrast one of \code{"ExIn"} or \code{"none"}. If \code{"ExIn"}
#'   (the default), genes that significantly differ between exonic and intronic changes
#'   are highlighted. \code{"none"} turns off gene highlighting.
#' @param minLfc \code{NULL} or \code{numeric(1)} with the minimal absolute log2
#'   fold change to color a gene. If \code{NULL} (the default), no fold changes
#'   are not used to select genes for highlighting.
#' @param maxFDR \code{numeric(1)} with maximal false discovery rate for gene
#'   highlighting.
#' @param genecolors Vector of length three specifying the colors to use for
#'   genes that are significantly up, down or unchanged.
#' @param ... further arguments past to \code{plot()}. Parameters that will be set
#'   automatically unless given in the arguments are:\describe{
#'     \item{pch}{: plot symbol (default: \code{"."})}
#'     \item{cex}{: plot symbol expansion factor (default: \code{2})}
#'     \item{col}{: plot symbol color (default: according to \code{contrast} and \code{genecolors})}
#'     \item{xlab/ylab}{: axis labels}
#'   }
#'
#' @return `NULL` (invisibly)
#'
#' @examples
#'   # see the help for runEISA() for a full example
#'
#' @importFrom graphics plot legend points
#'
#' @export
plotEISA <- function(x, contrast = c("ExIn", "none"),
                     minLfc = NULL, maxFDR = 0.05,
                     genecolors = c("#E41A1C", "#497AB3", "#222222"), ...) {
    # check arguments
    contrast <- match.arg(contrast)
    contrastName <- ifelse("contrastName" %in% names(x),
                           paste0(" (", x$contrastName, ")"), "")
    sigtab <- switch(contrast, ExIn = x$tab.ExIn, none = data.frame())
    if (nrow(sigtab) == 0L && contrast != "none")
        stop("'x' does not contain the requested statistics and can only ",
             "be plotted using contrast = 'none'. Note that at least two ",
             "replicates per condition are required to run the statistical ",
             "testing.")
    if (is.null(minLfc))
        minLfc <- 0.0
    stopifnot(exprs = {
        is.numeric(minLfc)
        length(minLfc) == 1L
        is.numeric(maxFDR)
        length(maxFDR) == 1L
        maxFDR >= 0.0
        maxFDR <= 1.0
        is.character(genecolors)
        length(genecolors) == 3L
    })

    # identify gene to highlight
    if (contrast == "none") {
        sig <- rep(FALSE, nrow(x$contrasts))
        sigDir <- 0.0
    } else {
        sig <- abs(sigtab$logFC) >= minLfc & sigtab$FDR <= maxFDR
        sigDir <- sign(sigtab$logFC[sig])
        message("identified ", sum(sig), " genes to highlight")
    }

    # set graphical parameters (user-defined colors take precedence)
    dotsL <- list(...)
    dotsL$x <- x$contrasts[, "Din"]
    dotsL$y <- x$contrasts[, "Dex"]
    if (!"pch" %in% names(dotsL))
        dotsL$pch <- "."
    if (!"cex" %in% names(dotsL))
        dotsL$cex <- 2L
    if (!"col" %in% names(dotsL))
        dotsL$col <- ifelse(sig,
                            ifelse(sigtab$logFC > 0.0,
                                   genecolors[1L], genecolors[2L]),
                            genecolors[3L])
    if (!"xlab" %in% names(dotsL))
        dotsL$xlab <- substitute(expression(paste(Delta, "intron", cn)),
                                 list(cn = contrastName))
    if (!"ylab" %in% names(dotsL))
        dotsL$ylab <- substitute(expression(paste(Delta, "exon", cn)),
                                 list(cn = contrastName))

    # Delta I vs. Delta E
    do.call(plot, dotsL)
    if (contrast != "none") {
        if (length(dotsL$col) == 1L)
            dotsL$col <- rep(dotsL$col, length(dotsL$x))
        points(dotsL$x[sig], dotsL$y[sig], pch = 20L, col = dotsL$col[sig])
        legend(x = "bottomright", bty = "n", pch = 20L, col = genecolors[c(1L, 2L)],
               legend = sprintf("%s (%d)",
                                c("Up", "Down"),
                                c(sum(sigDir == 1.0), sum(sigDir == -1.0))))
    }

    return(invisible(NULL))
}

#' Generate a transcript-to-gene mapping from a GRangesList
#'
#' Generate a \code{data.frame} mapping transcript IDs to gene IDs, based on
#' a GRangesList object generated by \code{getFeatureRanges}.
#'
#' @param grl \code{GRangesList} object, typically generated by
#'   \code{getFeatureRanges}
#' @param filepath Either \code{NULL} or the path to a file where the
#'   transcript-to-gene mapping will be written
#'
#' @author Charlotte Soneson
#'
#' @export
#'
#' @return Invisibly returns a \code{data.frame} with the transcript-to-gene
#'   mapping.
#'
#' @importFrom S4Vectors mcols
#' @importFrom BiocGenerics unlist subset
#' @importFrom utils write.table
#'
#' @examples
#'   ## Get feature ranges
#'   grl <- getFeatureRanges(
#'     gtf = system.file("extdata/small_example.gtf", package = "eisaR"),
#'     featureType = c("spliced", "intron"),
#'     intronType = "separate",
#'     flankLength = 5L,
#'     joinOverlappingIntrons = FALSE,
#'     verbose = TRUE
#'   )
#'
#'   ## Get transcript-to-gene mapping
#'   t2g <- getTx2Gene(grl = grl)
#'   t2g
#'
getTx2Gene <- function(grl, filepath = NULL) {
    grl <- BiocGenerics::unlist(grl)
    grl <- grl[grl$type == "exon"]
    t2g <- unique(data.frame(
        S4Vectors::mcols(grl)[, c("transcript_id", "gene_id")],
        stringsAsFactors = FALSE
    ))

    if (!is.null(filepath)) {
        utils::write.table(t2g, file = filepath, sep = "\t", quote = FALSE,
                           row.names = FALSE, col.names = FALSE)
    }
    invisible(t2g)
}

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)

## ----install_eisaR, eval=FALSE------------------------------------------------
# # BiocManager is needed to install Bioconductor packages
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# # Install eisaR
# BiocManager::install("eisaR")

## ----availableOnline, eval=FALSE----------------------------------------------
# pkgs <- c(BiocManager::available("TxDb"),
#           BiocManager::available("EnsDb"))

## ----annotation, message=FALSE------------------------------------------------
# load package
library(eisaR)

# get TxDb object
txdbFile <- system.file("extdata", "hg19sub.sqlite", package = "eisaR")
txdb <- AnnotationDbi::loadDb(txdbFile)

## ----regions------------------------------------------------------------------
# extract filtered exonic and gene body regions
regS <- getRegionsFromTxDb(txdb = txdb, strandedData = TRUE)
regU <- getRegionsFromTxDb(txdb = txdb, strandedData = FALSE)

lengths(regS)
lengths(regU)

regS$exons

## ----exportregions------------------------------------------------------------
library(rtracklayer)
export(regS$exons, "hg19sub_exons_stranded.gtf")
export(regS$genebodies, "hg19sub_genebodies_stranded.gtf")

## ----extdata------------------------------------------------------------------
library(QuasR)
file.copy(system.file(package = "QuasR", "extdata"), ".", recursive = TRUE)

## ----align--------------------------------------------------------------------
sampleFile <- file.path("extdata", "samples_rna_single.txt")
## Display the structure of the sampleFile
read.delim(sampleFile)

## Perform the alignment
proj <- qAlign(sampleFile = sampleFile,
               genome = file.path("extdata", "hg19sub.fa"),
               aligner = "Rhisat2", splicedAlignment = TRUE)
alignmentStats(proj)

## ----count--------------------------------------------------------------------
cntEx <- qCount(proj, regU$exons, orientation = "any")
cntGb <- qCount(proj, regU$genebodies, orientation = "any")
cntIn <- cntGb - cntEx
cntEx
cntIn

## ----loadcounts---------------------------------------------------------------
cntEx <- readRDS(system.file("extdata",
                             "Fig3abc_GSE33252_rawcounts_exonic.rds",
                             package = "eisaR"))
cntIn <- readRDS(system.file("extdata",
                             "Fig3abc_GSE33252_rawcounts_intronic.rds",
                             package = "eisaR"))

## ----runEISA------------------------------------------------------------------
# remove "width" column
rEx <- cntEx[, colnames(cntEx) != "width"]
rIn <- cntIn[, colnames(cntIn) != "width"]

# create condition factor (contrast will be TN - ES)
cond <- factor(c("ES", "ES", "TN", "TN"))

# run EISA
res <- runEISA(rEx, rIn, cond)

## ----compare------------------------------------------------------------------
res1 <- runEISA(rEx, rIn, cond, method = "Gaidatzis2015")
res2 <- runEISA(rEx, rIn, cond)

# number of quantifiable genes
nrow(res1$DGEList)
nrow(res2$DGEList)

# number of genes with significant post-transcriptional regulation
sum(res1$tab.ExIn$FDR < 0.05)
sum(res2$tab.ExIn$FDR < 0.05)

# method="Gaidatzis2015" results contain most of default results
summary(rownames(res2$contrasts)[res2$tab.ExIn$FDR < 0.05] %in%
        rownames(res1$contrasts)[res1$tab.ExIn$FDR < 0.05])

# comparison of deltas
ids <- intersect(rownames(res1$DGEList), rownames(res2$DGEList))
cor(res1$contrasts[ids, "Dex"], res2$contrasts[ids, "Dex"])
cor(res1$contrasts[ids, "Din"], res2$contrasts[ids, "Din"])
cor(res1$contrasts[ids, "Dex.Din"], res2$contrasts[ids, "Dex.Din"])
plot(res1$contrasts[ids, "Dex.Din"], res2$contrasts[ids, "Dex.Din"], pch = "*",
     xlab = expression(paste(Delta, "exon", -Delta, "intron for method='Gaidatzis2015'")),
     ylab = expression(paste(Delta, "exon", -Delta, "intron for default parameters")))

## ----modelSamples-------------------------------------------------------------
res3 <- runEISA(rEx, rIn, cond, modelSamples = FALSE)
res4 <- runEISA(rEx, rIn, cond, modelSamples = TRUE)
ids <- intersect(rownames(res3$contrasts), rownames(res4$contrasts))

# number of genes with significant post-transcriptional regulation
sum(res3$tab.ExIn$FDR < 0.05)
sum(res4$tab.ExIn$FDR < 0.05)

# modelSamples=TRUE results are a super-set of
# modelSamples=FALSE results
summary(rownames(res3$contrasts)[res3$tab.ExIn$FDR < 0.05] %in%
        rownames(res4$contrasts)[res4$tab.ExIn$FDR < 0.05])

# comparison of contrasts
diag(cor(res3$contrasts[ids, ], res4$contrasts[ids, ]))
plot(res3$contrasts[ids, 3L], res4$contrasts[ids, 3L], pch = "*",
     xlab = "Interaction effects for modelSamples=FALSE",
     ylab = "Interaction effects for modelSamples=TRUE")

# comparison of interaction significance
plot(-log10(res3$tab.ExIn[ids, "FDR"]), -log10(res4$tab.ExIn[ids, "FDR"]), pch = "*",
     xlab = "-log10(FDR) for modelSamples=FALSE",
     ylab = "-log10(FDR) for modelSamples=TRUE")
abline(a = 0.0, b = 1.0, col = "gray")
legend("bottomright", "y = x", bty = "n", lty = 1L, col = "gray")

## ----plotEISA-----------------------------------------------------------------
plotEISA(res)

## ----normalization------------------------------------------------------------
# remove column "width"
rEx <- cntEx[, colnames(cntEx) != "width"]
rIn <- cntIn[, colnames(cntIn) != "width"]
rAll <- rEx + rIn
fracIn <- colSums(rIn) / colSums(rAll)
summary(fracIn)

# scale counts to the mean library size,
# separately for exons and introns
normEx <- t(t(rEx) / colSums(rEx) * mean(colSums(rEx)))
normIn <- t(t(rIn) / colSums(rIn) * mean(colSums(rIn)))

# log transform (add a pseudocount of 8)
lognormEx <- log2(normEx + 8L)
lognormIn <- log2(normIn + 8L)

## ----quantgenes---------------------------------------------------------------
quantGenes <- rownames(rEx)[rowMeans(lognormEx) > 5.0 & rowMeans(lognormIn) > 5.0]
length(quantGenes)

## ----dIdE---------------------------------------------------------------------
Dex <- lognormEx[, c("MmTN_RNA_total_a", "MmTN_RNA_total_b")] -
    lognormEx[, c("MmES_RNA_total_a", "MmES_RNA_total_b")]
Din <- lognormIn[, c("MmTN_RNA_total_a", "MmTN_RNA_total_b")] -
    lognormIn[, c("MmES_RNA_total_a", "MmES_RNA_total_b")]
Dex.Din <- Dex - Din

cor(Dex[quantGenes, 1L], Dex[quantGenes, 2L])
cor(Din[quantGenes, 1L], Din[quantGenes, 2L])
cor(Dex.Din[quantGenes, 1L], Dex.Din[quantGenes, 2L])

## ----sig----------------------------------------------------------------------
# create DGEList object with exonic and intronic counts
library(edgeR)
cnt <- data.frame(Ex = rEx, In = rIn)
y <- DGEList(counts = cnt, genes = data.frame(ENTREZID = rownames(cnt)))

# select quantifiable genes and normalize
y <- y[quantGenes, ]
y <- calcNormFactors(y)

# design matrix with interaction term
region <- factor(c("ex", "ex", "ex", "ex", "in", "in", "in", "in"),
                 levels = c("in", "ex"))
cond <- rep(factor(c("ES", "ES", "TN", "TN")), 2L)
design <- model.matrix(~ region * cond)
rownames(design) <- colnames(cnt)
design

# estimate model parameters
y <- estimateDisp(y, design)
fit <- glmFit(y, design)

# calculate likelihood-ratio between full and reduced models
lrt <- glmLRT(fit)

# create results table
tt <- topTags(lrt, n = nrow(y), sort.by = "none")
head(tt$table[order(tt$table$FDR, decreasing = FALSE), ])

## ----plot---------------------------------------------------------------------
sig <- tt$table$FDR < 0.05
sum(sig)
sigDir <- sign(tt$table$logFC[sig])
cols <- ifelse(sig,
               ifelse(tt$table$logFC > 0.0, "#E41A1CFF", "#497AB3FF"),
               "#22222244")

# volcano plot
plot(tt$table$logFC, -log10(tt$table$FDR), col = cols, pch = 20L,
     xlab = expression(paste("RNA change (log2 ", Delta, "exon/", Delta, "intron)")),
     ylab = "Significance (-log10 FDR)")
abline(h = -log10(0.05), lty = 2L)
abline(v = 0.0, lty = 2L)
ppar <- par("usr")
text(x = ppar[1L] + 3.0 * par("cxy")[1L],
     y = ppar[4L], adj = c(0.0, 1.0),
     labels = sprintf("n=%d", sum(sigDir == -1L)), col = "#497AB3FF")
text(x = ppar[2L] - 3.0 * par("cxy")[1L],
     y = ppar[4L], adj = c(1.0, 1.0),
     labels = sprintf("n=%d", sum(sigDir ==  1L)), col = "#E41A1CFF")

# Delta I vs. Delta E
plot(rowMeans(Din)[quantGenes], rowMeans(Dex)[quantGenes], pch = 20L, col = cols,
     xlab = expression(paste(Delta, "intron (log2 TN/ES)")),
     ylab = expression(paste(Delta, "exon (log2 TN/ES)")))
legend(x = "bottomright", bty = "n", pch = 20L, col = c("#E41A1CFF", "#497AB3FF"),
       legend = sprintf("%s (%d)", c("Up", "Down"), c(sum(sigDir == 1L), sum(sigDir == -1L))))

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

## Genome
genome <- Biostrings::DNAStringSet(
    c(chr1 = "TTAACATTCGCTGGGGGAGATGACGAGACTAGCCGCCGCGTGGTCCTGCCGCATTATACGTGTTCAAGCGCCTACGTGGGTTGGGCAACCCGTGCCTATGGAGGCATGGACAAATTAGGTTCAACTTCAGCTACGTACGAGACCTAGAGGTAATAAGGGTATTTTACTCGGAGCATGTTTCAGTACGAACGTTAGATATC",
      chr2 = "CTATCGAAGTGGAATCTTGAAGAGCCCATCGGTTAAGGTCTCTCCAATGTCCAGCCTATTCTATGGCACGGCAGACCCGTTGTGCATCCACAGTGATAACTTACTTGGGCTCTTAATAGAGGAGTGTTGCCATTTTATCGGCTTGCACTCCAATTAGCACCAAGTGCCGTTATTGGGGTATTGCACTCATCAATAGCGTG")
)
genomeRevcompl <- Biostrings::reverseComplement(genome)

## Transcript-to-gene mapping
t2g <- data.frame(
    transcript_id = c("tx1.1", "tx1.2", "tx1.3", "tx2.1-I", "tx2.2-U", "tx3.1", "tx3.2", "tx3.3"),
    gene_id = c("g1", "g1", "g1", "g2-I", "g2-I", "g3-U", "g3-U", "g3-U"),
    stringsAsFactors = FALSE
)

## Transcripts
txome <- Biostrings::DNAStringSet(
    c(tx1.1 = paste0(substr(genome[["chr1"]], 116L, 120L),
                     substr(genome[["chr1"]], 141L, 160L)),
      tx1.2 = paste0(substr(genome[["chr1"]], 61L, 80L),
                     substr(genome[["chr1"]], 101L, 120L),
                     substr(genome[["chr1"]], 141L, 160L)),
      tx1.3 = paste0(substr(genome[["chr1"]], 61L, 90L)),
      `tx2.1-I` = paste0(substr(genomeRevcompl[["chr1"]], 131L, 150L),
                         substr(genomeRevcompl[["chr1"]], 171L, 190L)),
      `tx2.2-U` = paste0(substr(genomeRevcompl[["chr1"]], 121L, 140L),
                         substr(genomeRevcompl[["chr1"]], 156L, 165L),
                         substr(genomeRevcompl[["chr1"]], 181L, 190L)),
      tx3.1 = paste0(substr(genome[["chr2"]], 61L, 90L),
                     substr(genome[["chr2"]], 141L, 160L)),
      tx3.2 = paste0(substr(genome[["chr2"]], 61L, 80L),
                     substr(genome[["chr2"]], 101L, 120L),
                     substr(genome[["chr2"]], 141L, 160L)),
      tx3.3 = paste0(substr(genome[["chr2"]], 61L, 90L))
    )
)

## GTF
gtf <- GenomicRanges::GRanges(
    seqnames = rep(c("chr1", "chr2"), c(18L, 10L)),
    ranges = IRanges::IRanges(
        start = c(116L, 141L, 61L, 101L, 141L, 61L, 116L, 61L, 61L, 61L,
                  51L, 11L, 61L, 36L, 11L, 11L, 11L, 11L,
                  61L, 141L, 61L, 101L, 141L, 61L, 61L, 61L, 61L, 61L),
        end = c(120L, 160L, 80L, 120L, 160L, 90L, 160L, 160L, 90L, 160L,
                70L, 30L, 80L, 45L, 20L, 70L, 80L, 80L,
                90L, 160L, 80L, 120L, 160L, 90L, 160L, 160L, 90L, 160L)
    ),
    strand = rep(c("+", "-", "+"), c(10L, 8L, 10L)),
    type = c(rep("exon", 6L), rep("transcript", 3L), "gene", rep("exon", 5L),
             rep("transcript", 2L), "gene", rep("exon", 6L),
             rep("transcript", 3L), "gene"),
    gene_id = rep(c("g1", "g2-I", "g3-U"), c(10L, 8L, 10L)),
    transcript_id = c(rep("tx1.1", 2L), rep("tx1.2", 3L), "tx1.3", "tx1.1",
                      "tx1.2", "tx1.3", NA, rep("tx2.1-I", 2L),
                      rep("tx2.2-U", 3L), "tx2.1-I", "tx2.2-U", NA,
                      rep("tx3.1", 2L), rep("tx3.2", 3L), "tx3.3", "tx3.1",
                      "tx3.2", "tx3.3", NA),
    exon_id = c("E1", "E2", "E3", "E4", "E5", "E6", NA, NA, NA, NA,
                "E7", "E8", "E9", "E10", "E11", NA, NA, NA,
                "E12", "E13", "E14", "E15", "E16", "E17", NA, NA, NA, NA)
)

rtracklayer::export(gtf, con = "small_example.gtf", format = "gtf")
Biostrings::writeXStringSet(txome, filepath = "small_example_txome.fa")
Biostrings::writeXStringSet(genome, filepath = "small_example_genome.fa")
write.table(t2g, file = "small_example_t2g.txt", row.names = FALSE,
            col.names = FALSE, quote = FALSE, sep = "\t")

---
title: "Using eisaR for Exon-Intron Split Analysis (EISA)"
author: "Michael Stadler"
date: "`r Sys.Date()`"
bibliography: refs.bib
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Using eisaR for Exon-Intron Split Analysis (EISA)}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
```

# Introduction
Exon-Intron Split Analysis has been described by @eisa. It consists of separately
quantifying exonic and intronic alignments in RNA-seq data, in order to measure
changes in mature RNA and pre-mRNA reads across different experimental conditions.
We have shown that this allows quantification of transcriptional and post-transcriptional
regulation of gene expression.

The `eisaR` package contains convenience functions to facilitate the steps in an
exon-intron split analysis, which consists of:
1. preparing the annotation (exonic and gene body coordinate ranges, section \@ref(annotation))
2. quantifying RNA-seq alignments in exons and introns (sections \@ref(align) and \@ref(count))
3. calculating and comparing exonic and intronic changes across conditions (section \@ref(convenient))
4. visualizing the results (section \@ref(plot))

For the steps 1. and 2. above, this vignette makes use of Bioconductor annotation and
the `r Biocpkg("QuasR")` package. It is also possible to obtain count tables for exons and
introns using some other pipeline or approach, and directly start with step 3.

# Installation
To install the `eisaR` package, start R and enter:

```{r install_eisaR, eval=FALSE}
# BiocManager is needed to install Bioconductor packages
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# Install eisaR
BiocManager::install("eisaR")
```


# Preparing the annotation{#annotation}
As mentioned, `eisaR` uses gene annotations from Bioconductor.
They are provided in the form of `TxDb` or `EnsDb` objects, e.g. via packages such as `r Biocpkg("TxDb.Mmusculus.UCSC.mm10.knownGene")` or `r Biocpkg("EnsDb.Hsapiens.v86")`.
You can see available annotations using the following code:

```{r availableOnline, eval=FALSE}
pkgs <- c(BiocManager::available("TxDb"),
          BiocManager::available("EnsDb"))
```

If you would like to use an alternative source of gene annotations, you might
still be able to use `eisaR` by first converting your annotations into a `TxDb`
or an `EnsDb` (for creating a `TxDb` see `makeTxDb` in the `r Biocpkg("txdbmaker")`
package, for creating an `EnsDb` see `makeEnsembldbPackage` in the `r Biocpkg("ensembldb")`
package).

For this example, `eisaR` contains a small `TxDb` to illustrate how regions are extracted.
We will load it from a file. Alternatively, the object would be loaded using `library(...)`,
for example using `library(TxDb.Mmusculus.UCSC.mm10.knownGene)`.

```{r annotation, message=FALSE}
# load package
library(eisaR)

# get TxDb object
txdbFile <- system.file("extdata", "hg19sub.sqlite", package = "eisaR")
txdb <- AnnotationDbi::loadDb(txdbFile)
```

Exon and gene body regions are then extracted from the `TxDb`:

```{r regions}
# extract filtered exonic and gene body regions
regS <- getRegionsFromTxDb(txdb = txdb, strandedData = TRUE)
regU <- getRegionsFromTxDb(txdb = txdb, strandedData = FALSE)

lengths(regS)
lengths(regU)

regS$exons
```

As you can see, the filtering procedure removes slightly more genes for unstranded data
(`strandedData = FALSE`), as overlapping genes cannot be discriminated even if
they reside on opposite strands.

You can also export the obtained regions into files. This may be useful if
you plan to align and/or quantify reads outside of R. For example, you can use
`r Biocpkg("rtracklayer")` to export the regions in `regS` into `.gtf` files:

```{r exportregions}
library(rtracklayer)
export(regS$exons, "hg19sub_exons_stranded.gtf")
export(regS$genebodies, "hg19sub_genebodies_stranded.gtf")
```

# Quantify RNA-seq alignments in exons and introns
For this example we will use the `r Biocpkg("QuasR")` package for indexing and
alignment of short reads, and a small RNA-seq dataset that is contained in that
package. As mentioned, it is also possible to align or also quantify your reads
using an alternative aligner/counter, and skip over these steps. For more
details about the syntax, we refer to the documentation and vignette of the
`r Biocpkg("QuasR")` package.


## Align reads{#align}
Let's first copy the sample data from the `r Biocpkg("QuasR")` package to the
current working directory, all contained in a folder named `extdata`:

```{r extdata}
library(QuasR)
file.copy(system.file(package = "QuasR", "extdata"), ".", recursive = TRUE)
```

We next align the reads to a mini-genome (fasta file `extdata/hg19sub.fa`) using
`qAlign`. The `sampleFile` specifies the samples we want to use, and the paths
to the respective fastq files.

```{r align}
sampleFile <- file.path("extdata", "samples_rna_single.txt")
## Display the structure of the sampleFile
read.delim(sampleFile)

## Perform the alignment
proj <- qAlign(sampleFile = sampleFile,
               genome = file.path("extdata", "hg19sub.fa"),
               aligner = "Rhisat2", splicedAlignment = TRUE)
alignmentStats(proj)
```

## Count alignments in exons and gene bodies{#count}
Alignments in exons and gene bodies can now be counted using `qCount` and the
`regU` that we have generated earlier (assuming that the data is unstranded).
Intronic counts can then be obtained from the difference between gene bodies and
exons:

```{r count}
cntEx <- qCount(proj, regU$exons, orientation = "any")
cntGb <- qCount(proj, regU$genebodies, orientation = "any")
cntIn <- cntGb - cntEx
cntEx
cntIn
```

As mentioned, both alignments and counts can also be obtained using alternative
approaches. It is required that the two resulting exon and intron count tables
have identical structure (genes in rows, samples in columns, the same order of
rows and columns in both tables).

## Load full count tables
The above example only contains very few genes. For the rest of the vignette,
we will use count tables from a real RNA-seq experiment that are provided in the
`eisaR` package. The counts correspond to the raw data used in Figure 3a of @eisa
and are also available online from the [supplementary material](https://fmicompbio.github.io/projects/EISA/EISA.html):

```{r loadcounts}
cntEx <- readRDS(system.file("extdata",
                             "Fig3abc_GSE33252_rawcounts_exonic.rds",
                             package = "eisaR"))
cntIn <- readRDS(system.file("extdata",
                             "Fig3abc_GSE33252_rawcounts_intronic.rds",
                             package = "eisaR"))
```

# Run EISA conveniently{#convenient}
All the further steps in exon-intron split analysis can now be performed using
a single function `runEISA`. If you prefer to perform the analysis step-by-step,
you can skip now to section \@ref(stepwise).

```{r runEISA}
# remove "width" column
rEx <- cntEx[, colnames(cntEx) != "width"]
rIn <- cntIn[, colnames(cntIn) != "width"]

# create condition factor (contrast will be TN - ES)
cond <- factor(c("ES", "ES", "TN", "TN"))

# run EISA
res <- runEISA(rEx, rIn, cond)
```

## Alternative implementations of EISA
There are six arguments in `runEISA` (`modelSamples`, `geneSelection`, `effects`,
`statFramework`, `pscnt` and `sizeFactor`) that control gene filtering,
calculation of contrasts and the statistical method used, summarized in the
bullet list below.

The default values of these arguments correspond to the currently recommended way
of running EISA. You can also run EISA exactly as it was described by @eisa, by
setting `method = "Gaidatzis2015"`. This will override the values of the six
other arguments and set them according to the published algorithm (as indicated
below).

* `modelSamples`: Account for individual samples in statistical model? Possible values are:
  - `FALSE` (`method="Gaidatzis2015"`): use a model of the form `~ condition * region`
  - `TRUE` (default): use a model adjusting for the baseline differences among samples, and with condition-specific region effects (similar to the model described in section 3.5 of the `r Biocpkg("edgeR")` user guide)

* `geneSelection`: How to select detected genes. Possible values are:
  - `"filterByExpr"` (default): First, counts are normalized using `edgeR::calcNormFactors`,
    treating intronic and exonic counts as individual samples. Then, the
    `edgeR::filterByExpr` function is used with default parameters to select
    quantifiable genes.
  - `"none"`: This will use all the genes provided in the count tables, assuming
    that an appropriate selection of quantifiable genes has already been done.
  - `"Gaidatzis2015"` (`method="Gaidatzis2015"`): First, intronic and exonic counts
    are linearly scaled to the mean library size (estimated as the sum of all intronic
    or exonic counts, respectively). Then, quantifiable genes are selected as the
    genes with counts `x` that fulfill `log2(x + 8) > 5` in both exons and introns.

* `statFramework`: The framework within `edgeR` that is used for the statistical analysis.
  Possible values are:
  - `"QLF"` (default): quasi-likelihood F-test using `edgeR::glmQLFit` and
    `edgeR::glmQLFTest`. This framework is highly recommended as it gives stricter
    error rate control by accounting for the uncertainty in dispersion estimation.
  - `"LRT"` (`method="Gaidatzis2015"`): likelihood ratio test using `edgeR::glmFit`
    and `edgeR::glmLRT`.

* `effects`: How the effects (log2 fold-changes) are calculated. Possible values are:
  - `"predFC"` (default): Fold-changes are calculated using the fitted model with
    `edgeR::predFC` and the value provided to `pscnt`. Please note that if a
    sample factor is included in the statistical model (`modelSamples=TRUE`),
    effects cannot be obtained from that model. In that case, effects are obtained
    from a simpler model without sample effects.
  - `"Gaidatzis2015"` (`method="Gaidatzis2015"`): Fold-changes are calculated
    using the formula `log2((x + pscnt)/(y + pscnt))`. If `pscnt` is not set to 8,
    `runEISA` will warn that this deviates from the method used in Gaidatzis et al., 2015.

* `pscnt`: The pseudocount that is added to normalized counts before log transformation.
  For `geneSelection="Gaidatzis2015"`, `pscnt` is used both in gene selection as well as
  in the calculation of log2 fold-changes. Otherwise, `pscnt` is only used in the calculation
  of log2 fold-changes in `edgeR::predFC(, prior.count = pscnt)`.

* `sizeFactor`: How size factors (TMM normalization factors and library sizes)
are calculated and used within `eisaR`:
  - `"exon"` (default): Size factors are calculated for exonic counts and
  reused for the corresponding intronic counts.
  - `"intron"`: Size factors are calculated for intronic counts and
  reused for the corresponding exonic counts.
  - `"individual"` (`method="Gaidatzis2015"`): Size factors are calculated
  independently for exonic and intronic counts.

While different values for these arguments typically yield similar results,
the defaults are often less stringent compared to `method="Gaidatzis2015"` when
selecting quantifiable genes, but more stringent when calling significant changes
(especially with low numbers of replicates).

Here is an illustration of how the results differ between `method="Gaidatzis2015"` and
the defaults:

```{r compare}
res1 <- runEISA(rEx, rIn, cond, method = "Gaidatzis2015")
res2 <- runEISA(rEx, rIn, cond)

# number of quantifiable genes
nrow(res1$DGEList)
nrow(res2$DGEList)

# number of genes with significant post-transcriptional regulation
sum(res1$tab.ExIn$FDR < 0.05)
sum(res2$tab.ExIn$FDR < 0.05)

# method="Gaidatzis2015" results contain most of default results
summary(rownames(res2$contrasts)[res2$tab.ExIn$FDR < 0.05] %in%
        rownames(res1$contrasts)[res1$tab.ExIn$FDR < 0.05])

# comparison of deltas
ids <- intersect(rownames(res1$DGEList), rownames(res2$DGEList))
cor(res1$contrasts[ids, "Dex"], res2$contrasts[ids, "Dex"])
cor(res1$contrasts[ids, "Din"], res2$contrasts[ids, "Din"])
cor(res1$contrasts[ids, "Dex.Din"], res2$contrasts[ids, "Dex.Din"])
plot(res1$contrasts[ids, "Dex.Din"], res2$contrasts[ids, "Dex.Din"], pch = "*",
     xlab = expression(paste(Delta, "exon", -Delta, "intron for method='Gaidatzis2015'")),
     ylab = expression(paste(Delta, "exon", -Delta, "intron for default parameters")))
```

## On the estimation of interactions in a split-plot design experiment
The calculation of the significance of interactions (here whether the fold-changes
differ between exonic or intronic data) is well defined for experimental designs
where all samples are independent from one another. Within EISA, this is not the
case (each sample yields two data points, one for exons and one for introns). That
results in a dependency between data points: If a sample is affected by a problem
in the experiment, it might at the same time give rise to outlier values in both
exonic and intronic counts.

In statistics, such an experimental design is often referred to as a split-plot
design, and a recommended way to analyze interactions in such experiments would
be to use a mixed effect model with the plot (in our case, the sample) as a random
effect. The disadvantage here however would be that available packages for mixed
effect models are not designed for count data, and we therefore use an alternative
approach to explicitly model the sample dependency, by introducing sample-specific
columns into the design matrix (for `modelSamples=TRUE`). That sample factor is
nested in the condition factor (no sample can belong to more than one condition).
Thus, we are in the situation described in section 3.5 ('Comparisons both between and
within subjects') of the `r Biocpkg("edgeR")` user guide, and we use the approach
described there to define a design matrix with sample-specific baseline effects
as well as condition-specific region effects.

This has no impact on the effects (the log2 fold-changes of `modelSamples=TRUE`
and `modelSamples=FALSE` are nearly identical). However, in the presence of sample effects,
`modelSamples=TRUE` increases the sensitivity of detecting genes with significant
interactions. Here is a comparison of the EISA results with and without accounting
for the sample in the model:

```{r modelSamples}
res3 <- runEISA(rEx, rIn, cond, modelSamples = FALSE)
res4 <- runEISA(rEx, rIn, cond, modelSamples = TRUE)
ids <- intersect(rownames(res3$contrasts), rownames(res4$contrasts))

# number of genes with significant post-transcriptional regulation
sum(res3$tab.ExIn$FDR < 0.05)
sum(res4$tab.ExIn$FDR < 0.05)

# modelSamples=TRUE results are a super-set of
# modelSamples=FALSE results
summary(rownames(res3$contrasts)[res3$tab.ExIn$FDR < 0.05] %in%
        rownames(res4$contrasts)[res4$tab.ExIn$FDR < 0.05])

# comparison of contrasts
diag(cor(res3$contrasts[ids, ], res4$contrasts[ids, ]))
plot(res3$contrasts[ids, 3L], res4$contrasts[ids, 3L], pch = "*",
     xlab = "Interaction effects for modelSamples=FALSE",
     ylab = "Interaction effects for modelSamples=TRUE")

# comparison of interaction significance
plot(-log10(res3$tab.ExIn[ids, "FDR"]), -log10(res4$tab.ExIn[ids, "FDR"]), pch = "*",
     xlab = "-log10(FDR) for modelSamples=FALSE",
     ylab = "-log10(FDR) for modelSamples=TRUE")
abline(a = 0.0, b = 1.0, col = "gray")
legend("bottomright", "y = x", bty = "n", lty = 1L, col = "gray")
```


# Visualize EISA results{#plot}
We can now visualize the results by plotting intronic changes versus exonic
changes (genes with significant interactions, which are likely to be post-transcriptionally
regulated, are color coded):
```{r plotEISA}
plotEISA(res)
```

# Run EISA step-by-step{#stepwise}
As an alternative to `runEISA` (section \@ref(convenient)) and `plotEISA`
(section \@ref(plot)) described above, you can also analyze the data step-by-step
as described in @eisa. This may be preferable to understand each
individual step and be able to access intermediate results.

The results obtained in this way are identical to what you get with
`runEISA(..., method = "Gaidatzis2015")`, so if you are happy with `runEISA` you can
skip the rest of the vignette.

## Normalization
Normalization is performed separately on exonic and intronic counts, assuming
that varying exon over intron ratios between samples are of technical origin.

```{r normalization}
# remove column "width"
rEx <- cntEx[, colnames(cntEx) != "width"]
rIn <- cntIn[, colnames(cntIn) != "width"]
rAll <- rEx + rIn
fracIn <- colSums(rIn) / colSums(rAll)
summary(fracIn)

# scale counts to the mean library size,
# separately for exons and introns
normEx <- t(t(rEx) / colSums(rEx) * mean(colSums(rEx)))
normIn <- t(t(rIn) / colSums(rIn) * mean(colSums(rIn)))

# log transform (add a pseudocount of 8)
lognormEx <- log2(normEx + 8L)
lognormIn <- log2(normIn + 8L)
```

## Identify quantifiable genes
Genes with very low counts in either exons or introns are better removed, as
they will be too noisy to yield useful results. We use here a fixed cut-off
on the normalized, log-transformed intron and exonic counts:
```{r quantgenes}
quantGenes <- rownames(rEx)[rowMeans(lognormEx) > 5.0 & rowMeans(lognormIn) > 5.0]
length(quantGenes)
```

## Calculate $\Delta I$, $\Delta E$ and $\Delta E - \Delta I$
The count tables were obtained from a total RNA-seq experiments in mouse
embryonic stem (MmES) cells and derived terminal neurons (MmTN), with two replicates
for each condition.

We will now calculate the changes between neurons and ES cells in introns ($\Delta I$),
in exons ($\Delta E$), and the difference between the two ($\Delta E - \Delta I$):

```{r dIdE}
Dex <- lognormEx[, c("MmTN_RNA_total_a", "MmTN_RNA_total_b")] -
    lognormEx[, c("MmES_RNA_total_a", "MmES_RNA_total_b")]
Din <- lognormIn[, c("MmTN_RNA_total_a", "MmTN_RNA_total_b")] -
    lognormIn[, c("MmES_RNA_total_a", "MmES_RNA_total_b")]
Dex.Din <- Dex - Din

cor(Dex[quantGenes, 1L], Dex[quantGenes, 2L])
cor(Din[quantGenes, 1L], Din[quantGenes, 2L])
cor(Dex.Din[quantGenes, 1L], Dex.Din[quantGenes, 2L])
```
Both exonic and intronic changes are correlated across replicates, and so are the
differences, indicating a reproducible signal for post-transcriptional regulation.

## Statistical analysis
Finally, we use the replicate measurement in the `r Biocpkg("edgeR")` framework to
calculate the significance of the changes:
```{r sig}
# create DGEList object with exonic and intronic counts
library(edgeR)
cnt <- data.frame(Ex = rEx, In = rIn)
y <- DGEList(counts = cnt, genes = data.frame(ENTREZID = rownames(cnt)))

# select quantifiable genes and normalize
y <- y[quantGenes, ]
y <- calcNormFactors(y)

# design matrix with interaction term
region <- factor(c("ex", "ex", "ex", "ex", "in", "in", "in", "in"),
                 levels = c("in", "ex"))
cond <- rep(factor(c("ES", "ES", "TN", "TN")), 2L)
design <- model.matrix(~ region * cond)
rownames(design) <- colnames(cnt)
design

# estimate model parameters
y <- estimateDisp(y, design)
fit <- glmFit(y, design)

# calculate likelihood-ratio between full and reduced models
lrt <- glmLRT(fit)

# create results table
tt <- topTags(lrt, n = nrow(y), sort.by = "none")
head(tt$table[order(tt$table$FDR, decreasing = FALSE), ])
```

## Visualize the results
Finally, we visualize the results by plotting intronic changes versus exonic
changes (genes with significant interactions, which are likely to be post-transcriptionally
regulated, are color coded):
```{r plot}
sig <- tt$table$FDR < 0.05
sum(sig)
sigDir <- sign(tt$table$logFC[sig])
cols <- ifelse(sig,
               ifelse(tt$table$logFC > 0.0, "#E41A1CFF", "#497AB3FF"),
               "#22222244")

# volcano plot
plot(tt$table$logFC, -log10(tt$table$FDR), col = cols, pch = 20L,
     xlab = expression(paste("RNA change (log2 ", Delta, "exon/", Delta, "intron)")),
     ylab = "Significance (-log10 FDR)")
abline(h = -log10(0.05), lty = 2L)
abline(v = 0.0, lty = 2L)
ppar <- par("usr")
text(x = ppar[1L] + 3.0 * par("cxy")[1L],
     y = ppar[4L], adj = c(0.0, 1.0),
     labels = sprintf("n=%d", sum(sigDir == -1L)), col = "#497AB3FF")
text(x = ppar[2L] - 3.0 * par("cxy")[1L],
     y = ppar[4L], adj = c(1.0, 1.0),
     labels = sprintf("n=%d", sum(sigDir ==  1L)), col = "#E41A1CFF")

# Delta I vs. Delta E
plot(rowMeans(Din)[quantGenes], rowMeans(Dex)[quantGenes], pch = 20L, col = cols,
     xlab = expression(paste(Delta, "intron (log2 TN/ES)")),
     ylab = expression(paste(Delta, "exon (log2 TN/ES)")))
legend(x = "bottomright", bty = "n", pch = 20L, col = c("#E41A1CFF", "#497AB3FF"),
       legend = sprintf("%s (%d)", c("Up", "Down"), c(sum(sigDir == 1L), sum(sigDir == -1L))))
```

# Session information
The output in this vignette was produced under:

```{r sessionInfo}
sessionInfo()
```

# References

---
title: "Generating reference files for spliced and unspliced abundance estimation with alignment-free methods"
author: "Charlotte Soneson"
date: "`r Sys.Date()`"
output:
  BiocStyle::html_document:
    toc_float: true
vignette: >
  %\VignetteIndexEntry{Generating reference files for spliced and unspliced abundance estimation with alignment-free methods}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options:
  chunk_output_type: console
---

```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>"
)
options(width = 70L)
```

# Introduction

In this vignette, we show how to prepare reference files for estimating
abundances of spliced and unspliced abundances with alignment-free methods
(e.g., [Salmon](https://salmon.readthedocs.io/en/latest/salmon.html),
[alevin](https://salmon.readthedocs.io/en/latest/alevin.html) or
[kallisto](https://pachterlab.github.io/kallisto/)).
Such abundances are used as input, e.g., for estimation of RNA velocity in
single-cell data.

```{r setup}
library(eisaR)
```

# Generate feature ranges

The first step is to generate a `GRangesList` object containing the genomic
ranges for the features of interest (spliced transcripts, and either unspliced
transcripts or intron sequences).
This is done using the `getFeatureRanges()` function, based on a reference GTF
file.
Here, we exemplify this with a small subset of a GTF file from
[Gencode release 28](https://www.gencodegenes.org/human/release_28.html).
We extract genomic ranges for spliced transcript and introns, where the latter
are defined for each transcript separately (using the same terminology as in
the `r Biocpkg("BUSpaRse")` package).
For each intron, a flanking sequence of 50 nt is added on each side to
accommodate reads mapping across an exon/intron boundary.

```{r}
gtf <- system.file("extdata", "gencode.v28.annotation.sub.gtf",
                   package = "eisaR")
grl <- getFeatureRanges(
    gtf = gtf,
    featureType = c("spliced", "intron"),
    intronType = "separate",
    flankLength = 50L,
    joinOverlappingIntrons = FALSE,
    verbose = TRUE
)
```

The output from `getFeatureRanges()` is a `GRangesList` object, with all
the features of interest (both spliced transcripts and introns):

```{r}
grl
```

The `metadata` slot of the `GRangesList` object contains a list of the feature
IDs for each type of feature:

```{r}
lapply(S4Vectors::metadata(grl)$featurelist, head)
```

As we can see, the intron IDs are identified by a `-I` suffix.
Each feature is further annotated to a gene ID.
For the intronic features, the corresponding gene ID also bears the `-I`
suffix appended to the original gene ID.
Having separate gene IDs for spliced transcripts and introns allows, for
example, joint quantification of spliced and unspliced versions of a gene
with alevin.
Adding a suffix rather than defining a completely new gene ID allows us to
easily match corresponding spliced and unspliced feature IDs.
To further simplify the latter, the metadata of the `GRangesList` object
returned by `getFeatureRanges()` contains `data.frame`s matching the
corresponding gene IDs (as well as transcript IDs, if unspliced transcripts
are extracted):

```{r}
head(S4Vectors::metadata(grl)$corrgene)
```

# Extract feature sequences

Once the genomic ranges of the features of interest are extracted, we can
obtain the nucleotide sequences using the `extractTranscriptSeqs()` function
from the `r Biocpkg("GenomicFeatures")` package.
In addition to the ranges, this requires the genome sequence.
This can be obtained, for example, from the corresponding BSgenome package,
or from an external FASTA file.

```{r}
suppressPackageStartupMessages({
    library(BSgenome.Hsapiens.UCSC.hg38)
})
seqs <- GenomicFeatures::extractTranscriptSeqs(
    x = BSgenome.Hsapiens.UCSC.hg38,
    transcripts = grl
)
seqs
```

The resulting `DNAStringSet` can be written to a FASTA file and used to
generate an index for alignment-free methods such as _Salmon_ and _kallisto_.

# Generate an expanded GTF file

In addition to extracting feature sequences, we can also export a GTF file
with the full set of features.
This is useful, for example, in order to generate a linked transcriptome for
later import of estimated abundances with `r Biocpkg("tximeta")`.

```{r}
exportToGtf(
    grl,
    filepath = file.path(tempdir(), "exported.gtf")
)
```

In the exported GTF file, each entry of `grl` will correspond to a "transcript"
feature, and each individual range corresponds to an "exon" feature.
In addition, each gene is represented as a "gene" feature.

# Generate a transcript-to-gene mapping

Finally, we can export a `data.frame` (or a tab-separated test file) providing
a conversion table between "transcript" and "gene" identifiers.
This is needed to aggregate the transcript-level abundance estimates from
alignment-free methods such as _Salmon_ and _kallisto_ to the gene level.

```{r}
df <- getTx2Gene(grl)
head(df)
tail(df)
```

# Session info

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getRegionsFromTxDb.R
\name{getRegionsFromTxDb}
\alias{getRegionsFromTxDb}
\title{Get exonic/gene body regions from a transcript database.}
\usage{
getRegionsFromTxDb(txdb, exonExt = 10L, strandedData = TRUE)
}
\arguments{
\item{txdb}{a \code{TxDb} or an \code{EnsDb} object with the transcript annotations.}

\item{exonExt}{\code{numeric} (default = 10L). Exonic ranges will be extended
on either side by this many nucleotides, in order to avoid "bleed-over" of
exonic alignments into adjacent intronic regions.}

\item{strandedData}{\code{logical(1)}. If \code{TRUE}, the RNA-seq data is
assumed to be strand-specific, and therefore only overlapping genes that
are on the same strand will be filtered out. If \code{FALSE}, also genes
overlapping on opposite strands will be filtered out.}
}
\value{
a \code{list} with elements "exons" and "genebodies", containing
named \code{GenomicRanges} objects with ranges for exons and gene bodies,
respectively.
}
\description{
From a transcript database package (\code{\link[GenomicFeatures:TxDb-class]{TxDb}}),
extract exonic and gene body ranges for use with EISA. These regions can
be used to quantify RNA-seq alignments in exons and gene bodies, respectively.
Intronic counts can then be obtained from the difference between gene bodies
and exonic region counts.
}
\details{
The exonic regions are generated as follows:
\enumerate{
\item extract exons by gene from the \code{txdb}
\item extend each exon by \code{exonExt}
\item combine overlapping exons within each gene
\item create gene body ranges from the most extreme exonic coordinates
\item filter out genes that have only a single exon (no intron), have exons
on more than a single chromosome or on both strands, or that
overlap other genes
}
}
\examples{
if (requireNamespace("AnnotationDbi", quietly = TRUE)) {
    txdb <- AnnotationDbi::loadDb(system.file("extdata", "hg19sub.sqlite", package = "eisaR"))
    regL <- getRegionsFromTxDb(txdb)
    lengths(regL)
}

}
\seealso{
\code{\link[GenomicFeatures:TxDb-class]{TxDb}} for details on
\code{TxDb} objects and the \code{txdbmaker} package for how to create them,
e.g. from \code{.gtf} files.
}
\author{
Michael Stadler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotEISA.R
\name{plotEISA}
\alias{plotEISA}
\title{Visualize the results from an exon-intron split analysis.}
\usage{
plotEISA(
  x,
  contrast = c("ExIn", "none"),
  minLfc = NULL,
  maxFDR = 0.05,
  genecolors = c("#E41A1C", "#497AB3", "#222222"),
  ...
)
}
\arguments{
\item{x}{\code{list} with EISA results, typically the return value from \code{\link{runEISA}}}

\item{contrast}{one of \code{"ExIn"} or \code{"none"}. If \code{"ExIn"}
(the default), genes that significantly differ between exonic and intronic changes
are highlighted. \code{"none"} turns off gene highlighting.}

\item{minLfc}{\code{NULL} or \code{numeric(1)} with the minimal absolute log2
fold change to color a gene. If \code{NULL} (the default), no fold changes
are not used to select genes for highlighting.}

\item{maxFDR}{\code{numeric(1)} with maximal false discovery rate for gene
highlighting.}

\item{genecolors}{Vector of length three specifying the colors to use for
genes that are significantly up, down or unchanged.}

\item{...}{further arguments past to \code{plot()}. Parameters that will be set
automatically unless given in the arguments are:\describe{
\item{pch}{: plot symbol (default: \code{"."})}
\item{cex}{: plot symbol expansion factor (default: \code{2})}
\item{col}{: plot symbol color (default: according to \code{contrast} and \code{genecolors})}
\item{xlab/ylab}{: axis labels}
}}
}
\value{
\code{NULL} (invisibly)
}
\description{
\code{plotEISA} takes the return value from \code{\link{runEISA}}
and generates a scatterplot of intronic versus exonic changes.
}
\examples{
  # see the help for runEISA() for a full example

}
\author{
Michael Stadler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getTx2Gene.R
\name{getTx2Gene}
\alias{getTx2Gene}
\title{Generate a transcript-to-gene mapping from a GRangesList}
\usage{
getTx2Gene(grl, filepath = NULL)
}
\arguments{
\item{grl}{\code{GRangesList} object, typically generated by
\code{getFeatureRanges}}

\item{filepath}{Either \code{NULL} or the path to a file where the
transcript-to-gene mapping will be written}
}
\value{
Invisibly returns a \code{data.frame} with the transcript-to-gene
mapping.
}
\description{
Generate a \code{data.frame} mapping transcript IDs to gene IDs, based on
a GRangesList object generated by \code{getFeatureRanges}.
}
\examples{
  ## Get feature ranges
  grl <- getFeatureRanges(
    gtf = system.file("extdata/small_example.gtf", package = "eisaR"),
    featureType = c("spliced", "intron"),
    intronType = "separate",
    flankLength = 5L,
    joinOverlappingIntrons = FALSE,
    verbose = TRUE
  )
  
  ## Get transcript-to-gene mapping
  t2g <- getTx2Gene(grl = grl)
  t2g
   
}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runEISA.R
\name{runEISA}
\alias{runEISA}
\title{Run Exon-Intron Split Analysis.}
\usage{
runEISA(
  cntEx,
  cntIn,
  cond,
  method = NULL,
  modelSamples = TRUE,
  geneSelection = c("filterByExpr", "none", "Gaidatzis2015"),
  statFramework = c("QLF", "LRT"),
  legacyQLF = FALSE,
  effects = c("predFC", "Gaidatzis2015"),
  pscnt = 2,
  sizeFactor = c("exon", "intron", "individual"),
  recalcNormFactAfterFilt = TRUE,
  recalcLibSizeAfterFilt = FALSE,
  ...
)
}
\arguments{
\item{cntEx}{Gene by sample \code{matrix} with exonic counts, OR a
\code{SummarizedExperiment} with two assays named \code{exon} and
\code{intron}, containing exonic and intronic counts, respectively. If
\code{cntEx} is a \code{SummarizedExperiment}, \code{cntIn} will be
disregarded.}

\item{cntIn}{Gene by sample \code{matrix} with intronic counts. Must have the
same structure as \code{cntEx} (same number and order of rows and columns)
if \code{cntEx} is a matrix. Will be disregarded if \code{cntEx} is a
\code{SummarizedExperiment}.}

\item{cond}{\code{numeric}, \code{character} or \code{factor} with two levels
that groups the samples (columns of \code{cntEx} and \code{cntIn}) into two
conditions. The contrast will be defined as secondLevel - firstLevel.}

\item{method}{One of \code{NULL} (the default) or \code{"Gaidatzis2015"}. If
\code{"Gaidatzis2015"}, gene filtering, statistical analysis and
calculation of contrasts is performed as described in Gaidatzis et al.
2015, and the statistical analysis is based on \code{\link[edgeR]{glmFit}}
and \code{\link[edgeR:glmFit]{glmLRT}}. This is done by setting the arguments
\code{modelSamples}, \code{geneSelection}, \code{effects}, \code{pscnt},
\code{statFramework}, \code{sizeFactor}, \code{recalcNormFactAfterFilt}
and \code{recalcLibSizeAfterFilt}
to appropriate values (see details), overriding the defaults or any value
passed to these arguments. If \code{NULL}, the default values of the arguments
will be used instead (recommended).}

\item{modelSamples}{Whether to include a sample identifier in the design matrix
of the statistical model. If \code{TRUE}, potential sample effects
that affect both exonic and intronic counts of that sample will be taken
into account, which could result in higher sensitivity (default: \code{TRUE}).}

\item{geneSelection}{Controls how to select quantifyable genes. One of the
following:\describe{
\item{\code{"filterByExpr"}: }{(default) First, counts are normalized using
\code{\link[edgeR]{calcNormFactors}}, treating intronic and exonic counts
as individual samples. Then, \code{\link[edgeR]{filterByExpr}} is used
with default parameters to select quantifyable genes.}
\item{\code{"none"}: }{This will use all the genes provided in the count
tables, assuming that an appropriate selection of quantifyable genes has
already been done.}
\item{\code{"Gaidatzis2015"}: }{First, intronic and exonic counts are
linearly scaled to the mean library size (estimated as the sum of all
intronic or exonic counts, respectively). Then, quantifyable genes are
selected as the genes with counts \code{x} that fulfill
\code{log2(x + 8) > 5} in both exons and introns.}
}}

\item{statFramework}{Selects the framework within \code{edgeR} that is used
for the statistical analysis. One of:\describe{
\item{\code{"QLF"}: }{(default) Quasi-likelihood F-test using
\code{\link[edgeR:glmQLFTest]{glmQLFit}} and \code{\link[edgeR]{glmQLFTest}}. This
framework is highly recommended as it gives stricter error rate control
by accounting for the uncertainty in dispersion estimation.}
\item{\code{"LRT"}: }{Likelihood ratio test using \code{\link[edgeR]{glmFit}}
and \code{\link[edgeR:glmFit]{glmLRT}}}.
}}

\item{legacyQLF}{Whether to use the 'legacy' version of
\code{\link[edgeR:glmQLFTest]{glmQLFit}}. See \code{\link[edgeR:glmQLFTest]{glmQLFit}}
for more details. If \code{FALSE}, the new method introduced in
\code{edgeR} 4.0.0 is used.}

\item{effects}{How the effects (contrasts or log2 fold-changes) are calculated.
One of:\describe{
\item{\code{"predFC"}: }{(default) Fold-changes are calculated using
the fitted model with \code{\link[edgeR]{predFC}} with
\code{prior.count = pscnt}. Please note that if a sample factor is
included in the model (\code{modelSamples=TRUE}), effects cannot be
obtained from that model. In that case, effects are obtained from a
simpler model without sample effects.}
\item{\code{"Gaidatzis2015"}: }{Fold-changes are calculated using the
formula \code{log2((x + pscnt)/(y + pscnt))}. If \code{pscnt} is not
set to 8, \code{runEISA} will warn that this deviates from the method
used in Gaidatzis et al., 2015.}
}}

\item{pscnt}{\code{numeric(1)} with pseudocount to add to read counts
(default: 2). For \code{method = "Gaidatzis2015"}, it is set to 8.
It is added to scaled read counts used in \code{geneSelection = "Gaidatzis2015"}
and \code{effects = "Gaidatzis2015"}, or else used in \code{cpm(..., prior.count = pscnt)}
and \code{predFC(..., prior.count = pscnt)}.}

\item{sizeFactor}{How the size factors are calculated in the analysis.
If 'exon' (default), the exon-derived size factors are used also for the
columns corresponding to intronic counts. If 'intron', the intron-derived
size factors are used also for the columns corresponding to exonic
counts. If 'individual', column-wise size factors are calculated.}

\item{recalcNormFactAfterFilt}{Logical, indicating whether normalization
factors should be recalculated after filtering out lowly
expressed genes.}

\item{recalcLibSizeAfterFilt}{Logical, indicating whether library
sizes should be recalculated after filtering out lowly
expressed genes.}

\item{...}{additional arguments passed to the \code{\link[edgeR]{DGEList}}
constructor, such as \code{lib.size} or \code{genes}.}
}
\value{
a \code{list} with elements \describe{
\item{fracIn}{fraction intronic counts in each sample}
\item{contrastName}{contrast name}
\item{contrasts}{contrast matrix for quantifyable genes, with average log2
fold-changes in exons (\code{Dex}), in introns (\code{Din}), and average
difference between log2 fold-changes in exons and introns (\code{Dex.Din})}
\item{DGEList}{\code{\link[edgeR]{DGEList}} object used in model fitting}
\item{tab.ExIn}{statisical results for differential changes between exonic
and intronic contrast, an indication for post-transcriptional regulation.}
\item{contr.ExIn}{contrast vector used for testing the difference between
exonic and intronic contrast (results in \code{tab.ExIn})}
\item{designMatrix}{design matrix used for testing the difference between
exonic and intronic contrast (results in \code{tab.ExIn})}
\item{params}{a \code{list} with parameter values used to run EISA}
}
}
\description{
Starting from count tables with exonic and intronic counts
for two conditions, perform all the steps in EISA (normalize, identify
quantifyable genes, calculate contrasts and their significance).
}
\details{
Setting \code{method = "Gaidatzis2015"} has precedence over other
argument values and corresponds to setting:
\code{modelSamples = FALSE, geneSelection = "Gaidatzis2015",
  statFramework = "LRT", effects = "Gaidatzis2015", pscnt = 8, 
  sizeFactor = "individual", recalcNormFactAfterFilt = TRUE, 
  recalcLibSizeAfterFilt = FALSE}.
}
\examples{
cntEx <- readRDS(system.file("extdata", "Fig3abc_GSE33252_rawcounts_exonic.rds",
                             package = "eisaR"))[,-1]
cntIn <- readRDS(system.file("extdata", "Fig3abc_GSE33252_rawcounts_intronic.rds",
                             package = "eisaR"))[,-1]
cond <- factor(c("ES","ES","TN","TN"))
res <- runEISA(cntEx, cntIn, cond)
plotEISA(res)

}
\references{
Analysis of intronic and exonic reads in RNA-seq data characterizes
transcriptional and post-transcriptional regulation.
Dimos Gaidatzis, Lukas Burger, Maria Florescu and Michael B. Stadler
Nature Biotechnology, 2015 Jul;33(7):722-9. doi: 10.1038/nbt.3269.
}
\seealso{
\code{\link[edgeR]{DGEList}} for \code{DGEList} object construction,
\code{\link[edgeR]{calcNormFactors}} for normalization,
\code{\link[edgeR]{filterByExpr}} for gene selection,
\code{\link[edgeR]{glmFit}} and \code{\link[edgeR:glmQLFTest]{glmQLFit}} for statistical
analysis.
}
\author{
Michael Stadler
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/exportToGtf.R
\name{exportToGtf}
\alias{exportToGtf}
\title{Export GRangesList to GTF}
\usage{
exportToGtf(grl, filepath)
}
\arguments{
\item{grl}{\code{GRangesList} object, typically generated by
\code{getFeatureRanges}}

\item{filepath}{Path to output GTF file}
}
\value{
Does not return anything, generates a GTF file
}
\description{
Export the features in a GRangesList generated by \code{getFeatureRanges}
to a GTF file. The function will represent each row of each of the
entries as an "exon", each individual entry as a "transcript", and
aggregate all features belonging to the same gene as a "gene" entry in
the GTF file.
}
\examples{
  ## Get feature ranges
  grl <- getFeatureRanges(
    gtf = system.file("extdata/small_example.gtf", package = "eisaR"),
    featureType = c("spliced", "intron"),
    intronType = "separate",
    flankLength = 5L,
    joinOverlappingIntrons = FALSE,
    verbose = TRUE
  )
  
  ## Export GTF
  exportToGtf(grl = grl, filepath = file.path(tempdir(), "exported.gtf"))

}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getFeatureRanges.R
\name{getFeatureRanges}
\alias{getFeatureRanges}
\title{Generate a GRangesList object with feature ranges}
\usage{
getFeatureRanges(
  gtf,
  featureType = c("spliced", "intron"),
  intronType = "separate",
  flankLength = 90L,
  joinOverlappingIntrons = FALSE,
  collapseIntronsByGene = FALSE,
  keepIntronInFeature = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{gtf}{Path to gtf file.}

\item{featureType}{Character vector indicating the type(s) of features to
extract, any subset of \code{c("spliced", "intron", "unspliced")}.}

\item{intronType}{Character vector indicating how to define the introns
(only used if "intron" is part of \code{featureType}). Has to be either
"separate" (introns are defined for each transcript separately) or
"collapse" (transcripts of the same gene are first collapsed before
introns are defined as any non-exonic part of the gene locus).}

\item{flankLength}{Integer scalar indicating the length of the flanking
sequence added to each side of each extracted intron (only used if
"intron" is included in \code{featureType}).}

\item{joinOverlappingIntrons}{Logical scalar indicating whether two introns
that overlap (after adding the flanking sequences) should be joined into
one feature.}

\item{collapseIntronsByGene}{Logical scalar indicating whether to collapse
overlapping intron ranges by gene after extraction.}

\item{keepIntronInFeature}{Logical scalar indicating whether introns
(after adding the flank length) should be restricted to stay within the
boundaries of the feature they were generated from. For backwards
compatibility, the default is \code{FALSE}.}

\item{verbose}{Logical scalar, whether to print out progress messages.}
}
\value{
Returns a \code{GRangesList} object where each element represents
one extracted feature. The metadata of this object contains two
\code{data.frame}s mapping corresponding identifiers between the
different feature types, as well as a list of all features for each type.
}
\description{
Generate a GRangesList object with genomic ranges for (any combination of)
spliced transcripts, unspliced transcripts and introns.
}
\examples{
  ## Get feature ranges
  grl <- getFeatureRanges(
    gtf = system.file("extdata/small_example.gtf", package = "eisaR"),
    featureType = c("spliced", "intron"),
    intronType = "separate",
    flankLength = 5L,
    joinOverlappingIntrons = FALSE,
    collapseIntronsByGene = FALSE,
    verbose = TRUE
  )
  
  ## GRangesList
  grl
  
  ## Corresponding transcript/gene IDs
  S4Vectors::metadata(grl)$corrtx
  S4Vectors::metadata(grl)$corrgene
  
  ## List of features of different types
  S4Vectors::metadata(grl)$featurelist
  
  ## Get feature sequences
  if (requireNamespace("BSgenome", quietly = TRUE) &&
      requireNamespace("GenomicFeatures", quietly = TRUE)) {
    library(BSgenome)
    genome <- Biostrings::readDNAStringSet(
      system.file("extdata/small_example_genome.fa", package = "eisaR"))
    seqs <- GenomicFeatures::extractTranscriptSeqs(x = genome, 
                                                   transcripts = grl)
    seqs
  }
 
}
\author{
Charlotte Soneson
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eisa-package.R
\docType{package}
\name{eisaR-package}
\alias{eisaR}
\alias{eisaR-package}
\title{eisaR: Exon-Intron Split Analysis (EISA) in R}
\description{
Exon-intron split analysis (EISA) uses ordinary RNA-seq data to measure changes in mature RNA and pre-mRNA reads across different experimental conditions to quantify transcriptional and post-transcriptional regulation of gene expression. For details see Gaidatzis et al., Nat Biotechnol 2015. doi: 10.1038/nbt.3269. eisaR implements the major steps of EISA in R.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/fmicompbio/eisaR}
  \item Report bugs at \url{https://github.com/fmicompbio/eisaR/issues}
}

}
\author{
\strong{Maintainer}: Michael Stadler \email{michael.stadler@fmi.ch}

Authors:
\itemize{
  \item Dimos Gaidatzis \email{dimosthenis.gaidatzis@fmi.ch}
  \item Lukas Burger \email{lukas.burger@fmi.ch}
  \item Charlotte Soneson \email{charlotte.soneson@fmi.ch}
}

}
\keyword{internal}
